import numpy
from pyscf import numpy as np
from pyscf.gto.mole import (ATOM_OF, ANG_OF, NPRIM_OF, NCTR_OF,
                            KAPPA_OF, PTR_EXP, PTR_COEFF, PTR_ENV_START)

def uncontract(mol):
    """
    Uncontract basis shell by shell
    (exponents not sorted, basis not normalized)
    """
    mol1 = mol.copy()
    tmp = []
    env = []
    bas = []
    ioff = istart = PTR_ENV_START + mol.natm * 4
    for i in range(len(mol._bas)):
        iatm = mol._bas[i,ATOM_OF]
        l = mol._bas[i,ANG_OF]
        nprim = mol._bas[i,NPRIM_OF]
        kappa = mol._bas[i,KAPPA_OF]
        ptr_exp = mol._bas[i,PTR_EXP]
        if ptr_exp not in tmp:
            tmp.append(ptr_exp)
            for j in range(nprim):
                env.append([mol._env[ptr_exp+j], 1.])
                bas.append([iatm, l, 1, 1, kappa, ioff, ioff+1, ptr_exp])
                ioff += 2

    env = numpy.asarray(env).flatten()
    mol1._env = numpy.hstack((mol._env[:istart], env))

    bas = numpy.asarray(bas)
    ptr_exp = mol._bas[:,PTR_EXP]
    _bas = []
    for i, ptr in enumerate(ptr_exp):
        iatm = mol._bas[i,ATOM_OF]
        bas_tmp = bas[numpy.where(bas[:,-1] == ptr)[0]]
        bas_tmp[:,ATOM_OF] = iatm
        _bas.append(bas_tmp)
    _bas = numpy.vstack(tuple(_bas))
    _bas[:,-1] = 0
    mol1._bas = _bas
    return mol1

def setup_exp(mol):
    tmp = []
    es = np.empty([0], dtype=float)
    _env_of = numpy.empty([0], dtype=numpy.int32)
    offset = 0
    es_of = []
    for i in range(len(mol._bas)):
        nprim = mol._bas[i,NPRIM_OF]
        ptr_exp = mol._bas[i,PTR_EXP]
        if ptr_exp not in tmp:
            tmp.append(ptr_exp)
            es = np.append(es, mol._env[ptr_exp : ptr_exp+nprim])
            _env_of = numpy.append(_env_of, numpy.arange(ptr_exp,ptr_exp+nprim))
            es_of.append(offset)
            offset += nprim
    tmp = numpy.asarray(tmp, dtype=numpy.int32)
    es_of = numpy.asarray(es_of, dtype=numpy.int32)
    ptr_exp = mol._bas[:,PTR_EXP]
    idx = []
    for ptr in ptr_exp:
        idx.append(numpy.where(ptr == tmp)[0])
    idx = numpy.asarray(idx).flatten()
    es_of = es_of[idx]
    return es, es_of, _env_of

def setup_ctr_coeff(mol):
    tmp = []
    cs = np.empty([0], dtype=float)
    _env_of = numpy.empty([0], dtype=numpy.int32)
    offset = 0
    cs_of = []
    for i in range(len(mol._bas)):
        nprim = mol._bas[i,NPRIM_OF]
        nctr = mol._bas[i,NCTR_OF]
        ptr_coeff = mol._bas[i,PTR_COEFF]
        if ptr_coeff not in tmp:
            tmp.append(ptr_coeff)
            cs = np.append(cs, mol._env[ptr_coeff : ptr_coeff+nprim*nctr])
            _env_of = numpy.append(_env_of, numpy.arange(ptr_coeff,ptr_coeff+nprim*nctr))
            cs_of.append(offset)
            offset += nprim*nctr
    tmp = numpy.asarray(tmp, dtype=numpy.int32)
    cs_of = numpy.asarray(cs_of, dtype=numpy.int32)
    ptr_coeff = mol._bas[:,PTR_COEFF]
    idx = []
    for ptr in ptr_coeff:
        idx.append(numpy.where(ptr == tmp)[0])
    idx = numpy.asarray(idx).flatten()
    cs_of = cs_of[idx]
    return cs, cs_of, _env_of



"""
Helper functions for jax
"""

import dataclasses
import numpy
import jax
from jax import tree_util
from pyscf import __config__

PYSCFAD = getattr(__config__, "pyscfad", False)

def stop_grad(x):
    if PYSCFAD:
        return jax.lax.stop_gradient(x)
    else:
        return x

def jit(fun, **kwargs):
    if PYSCFAD:
        return jax.jit(fun, **kwargs)
    else:
        return fun

def vmap_numpy(fun, in_axes=0, out_axes=0, axis_name=None, axis_size=None, signature=None):
    if axis_name is not None:
        raise NotImplementedError
    if axis_size is not None:
        raise NotImplementedError
    if not isinstance(out_axes, int):
        raise NotImplementedError

    def vmap_f(*args):
        if isinstance(in_axes, int):
            in_axes_loc = (in_axes,) * len(args)
        else:
            in_axes_loc = in_axes

        if isinstance(in_axes_loc, (list, tuple)):
            excluded = []
            vmap_args = []
            assert len(in_axes_loc) == len(args)
            for i, axis in enumerate(in_axes_loc):
                if axis is None:
                    excluded.append(i)
                    vmap_args.append(args[i])
                elif isinstance(axis, int):
                    vmap_args.append(numpy.moveaxis(args[i], axis, 0))
                else:
                    raise KeyError
            if len(excluded) > 0:
                excluded = set(excluded)
            else:
                excluded = None

            vfun = numpy.vectorize(fun, excluded=excluded, signature=signature)
            out = vfun(*vmap_args)
        else:
            raise KeyError

        if out_axes != 0:
            out = numpy.moveaxis(out, 0, out_axes)
        return out

    return vmap_f

if PYSCFAD:
    def vmap(fun, in_axes=0, out_axes=0, axis_name=None, axis_size=None, signature=None):
        f_vmap = jax.vmap(fun, in_axes=in_axes, out_axes=out_axes,
                          axis_name=axis_name, axis_size=axis_size)
        return f_vmap
else:
    vmap = vmap_numpy

if PYSCFAD:
    custom_jvp = jax.custom_jvp
else:
    class custom_jvp():
        '''
        A fake custom_jvp that does nothing
        '''
        def __init__(self, fun, *args, **kwargs):
            self.fun = fun
            self.jvp = None

        def defjvp(self, jvp):
            self.jvp = jvp
            return jvp

        def __call__(self, *args, **kwargs):
            return self.fun(*args, **kwargs)


def dataclass(cls):
    data_cls = dataclasses.dataclass()(cls)
    data_fields = []
    meta_fields = []
    for field_name, field_info in data_cls.__dataclass_fields__.items():
        is_pytree_node = field_info.metadata.get('pytree_node', False)
        if is_pytree_node:
            data_fields.append(field_name)
        else:
            meta_fields.append(field_name)

    def tree_flatten(obj):
        children =  tuple(getattr(obj, key, None) for key in data_fields)
        metadata =  tuple(getattr(obj, key, None) for key in meta_fields)
        return children, metadata

    def tree_unflatten(metadata, children):
        data_args = tuple(zip(data_fields, children))
        meta_args = tuple(zip(meta_fields, metadata))
        kwargs = dict(data_args + meta_args)
        obj = data_cls(**kwargs)
        return obj

    tree_util.register_pytree_node(data_cls,
                                   tree_flatten,
                                   tree_unflatten)
    return data_cls

def field(pytree_node=False, **kwargs):
    return dataclasses.field(metadata={'pytree_node': pytree_node}, **kwargs)



from distutils.version import LooseVersion
import jax
import jax.ops
from jax import numpy as jnp
from jax.config import config as jax_config
jax_config.update("jax_enable_x64", True)

# pylint: disable=no-member

index = jnp.index_exp
if LooseVersion(jax.__version__) < '0.2.22':
    index_update = jax.ops.index_update
    index_add = jax.ops.index_add
    index_mul = jax.ops.index_mul
else:
    def index_update(x, idx, y, indices_are_sorted=False, unique_indices=False):
        x = jnp.asarray(x)
        y = jnp.asarray(y)
        return x.at[idx].set(y)

    def index_add(x, idx, y, indices_are_sorted=False, unique_indices=False):
        x = jnp.asarray(x)
        y = jnp.asarray(y)
        return x.at[idx].add(y)

    def index_mul(x, idx, y, indices_are_sorted=False, unique_indices=False):
        x = jnp.asarray(x)
        y = jnp.asarray(y)
        return x.at[idx].multiply(y)


del LooseVersion



from pyscfad.pbc.dft import rks
from pyscfad.pbc.dft import krks

def RKS(cell, *args, **kwargs):
    return rks.RKS(cell, *args, **kwargs)

def KRKS(cell, *args, **kwargs):
    return krks.KRKS(cell, *args, **kwargs)



import numpy
from pyscf import __config__
from pyscf.lib import logger
from pyscf.pbc.dft import krks as pyscf_krks
from pyscf.pbc.dft import gen_grid, multigrid
from pyscfad import util
from pyscfad.lib import numpy as np
from pyscfad.lib import stop_grad
from pyscfad.dft.rks import VXC
from pyscfad.pbc.scf import khf
from pyscfad.pbc.dft import rks

def get_veff(ks, cell=None, dm=None, dm_last=0, vhf_last=0, hermi=1,
             kpts=None, kpts_band=None):
    if cell is None: cell = ks.cell
    if dm is None: dm = ks.make_rdm1()
    if kpts is None: kpts = ks.kpts

    log = logger.new_logger(ks)

    omega, alpha, hyb = ks._numint.rsh_and_hybrid_coeff(ks.xc, spin=cell.spin)
    hybrid = abs(hyb) > 1e-10 or abs(alpha) > 1e-10

    if not hybrid and isinstance(ks.with_df, multigrid.MultiGridFFTDF):
        n, exc, vxc = multigrid.nr_rks(ks.with_df, ks.xc, dm, hermi,
                                       kpts, kpts_band,
                                       with_j=True, return_j=False)
        log.debug('nelec by numeric integration = %s', n)
        log.timer('vxc')
        return vxc

    # ndim = 3 : dm.shape = (nkpts, nao, nao)
    ground_state = (getattr(dm, 'ndim', None) == 3 and 
                    kpts_band is None)

# For UniformGrids, grids.coords does not indicate whehter grids are initialized
    if ks.grids.non0tab is None:
        ks.grids.build(with_non0tab=True)
        if (isinstance(ks.grids, gen_grid.BeckeGrids) and
            ks.small_rho_cutoff > 1e-20 and ground_state):
            ks.grids = rks.prune_small_rho_grids_(ks, stop_grad(cell), stop_grad(dm), ks.grids, kpts)
        log.timer('setting up grids')

    if hermi == 2:  # because rho = 0
        n, exc, vxc = 0, 0, 0
    else:
        n, exc, vxc = ks._numint.nr_rks(cell, ks.grids, ks.xc, dm, 0,
                                        kpts, kpts_band)
        log.debug('nelec by numeric integration = %s', n)
        log.timer('vxc')

    weight = 1./len(kpts)
    if not hybrid:
        vj = ks.get_j(cell, dm, hermi, kpts, kpts_band)
        vxc += vj
    else:
        if getattr(ks.with_df, '_j_only', False):  # for GDF and MDF
            ks.with_df._j_only = False
        vj, vk = ks.get_jk(cell, dm, hermi, kpts, kpts_band)
        vk *= hyb
        if abs(omega) > 1e-10:
            vklr = ks.get_k(cell, dm, hermi, kpts, kpts_band, omega=omega)
            vklr *= (alpha - hyb)
            vk += vklr
        vxc += vj - vk * .5

        if ground_state:
            exc -= np.einsum('Kij,Kji', dm, vk).real * .5 * .5 * weight

    if ground_state:
        ecoul = np.einsum('Kij,Kji', dm, vj).real * .5 * weight
    else:
        ecoul = None

    vxc = VXC(vxc=vxc, ecoul=ecoul, exc=exc, vj=None, vk=None)
    del log
    return vxc

@util.pytree_node(khf.Traced_Attributes, num_args=1)
class KRKS(rks.KohnShamDFT, khf.KRHF):
    def __init__(self, cell, kpts=numpy.zeros((1,3)), xc='LDA,VWN',
                 exxdiv=getattr(__config__, 'pbc_scf_SCF_exxdiv', 'ewald'), **kwargs):
        khf.KRHF.__init__(self, cell, kpts, exxdiv, **kwargs)
        rks.KohnShamDFT.__init__(self, xc)
        self.__dict__.update(kwargs)
        # NOTE this has to be after __dict__ update,
        # otherwise stop_grad(mol) won't work.
        # Currently, no grid response is considered.
        rks.KohnShamDFT.__post_init__(self)


    def dump_flags(self, verbose=None):
        khf.KRHF.dump_flags(self, verbose)
        rks.KohnShamDFT.dump_flags(self, verbose)
        return self

    get_veff = get_veff

    def energy_elec(self, dm_kpts=None, h1e_kpts=None, vhf=None):
        if h1e_kpts is None: h1e_kpts = self.get_hcore(self.cell, self.kpts)
        if dm_kpts is None: dm_kpts = self.make_rdm1()
        if vhf is None or getattr(vhf, 'ecoul', None) is None:
            vhf = self.get_veff(self.cell, dm_kpts)

        weight = 1./len(h1e_kpts)
        e1 = weight * np.einsum('kij,kji', h1e_kpts, dm_kpts)
        tot_e = e1 + vhf.ecoul + vhf.exc
        self.scf_summary['e1'] = stop_grad(e1.real)
        self.scf_summary['coul'] = stop_grad(vhf.ecoul.real)
        self.scf_summary['exc'] = stop_grad(vhf.exc.real)
        logger.debug(self, 'E1 = %s  Ecoul = %s  Exc = %s', e1, vhf.ecoul, vhf.exc)
        return tot_e.real, vhf.ecoul + vhf.exc

    get_rho = pyscf_krks.get_rho



import numpy as np
from pyscfad.lib import numpy as jnp

def fft(f, mesh):
    '''
    3D FFT with jax.numpy backend
    '''
    if f.size == 0:
        return np.zeros_like(f)

    f3d = f.reshape(-1, *mesh)
    assert(f3d.shape[0] == 1 or f[0].size == f3d[0].size)
    g3d = jnp.fft.fftn(f3d, axes=(1,2,3))
    ngrids = np.prod(mesh)
    if f.ndim == 1 or (f.ndim == 3 and f.size == ngrids):
        return g3d.ravel()
    else:
        return g3d.reshape(-1, ngrids)

def ifft(g, mesh):
    '''
    3D inverse FFT with jax.numpy backend
    '''
    if g.size == 0:
        return np.zeros_like(g)

    g3d = g.reshape(-1, *mesh)
    assert(g3d.shape[0] == 1 or g[0].size == g3d[0].size)
    f3d = jnp.fft.ifftn(g3d, axes=(1,2,3))
    ngrids = np.prod(mesh)
    if g.ndim == 1 or (g.ndim == 3 and g.size == ngrids):
        return f3d.ravel()
    else:
        return f3d.reshape(-1, ngrids)

def fftk(f, mesh, expmikr):
    r'''Perform the 3D FFT of a real-space function which is (periodic*e^{ikr}).

    fk(k+G) = \sum_r fk(r) e^{-i(k+G)r} = \sum_r [f(k)e^{-ikr}] e^{-iGr}
    '''
    return fft(f*expmikr, mesh)


def ifftk(g, mesh, expikr):
    r'''Perform the 3D inverse FFT of f(k+G) into a function which is (periodic*e^{ikr}).

    fk(r) = (1/Ng) \sum_G fk(k+G) e^{i(k+G)r} = (1/Ng) \sum_G [fk(k+G)e^{iGr}] e^{ikr}
    '''
    return ifft(g, mesh) * expikr



from functools import partial
import numpy
from pyscf.gto.moleintor import make_loc
from pyscf.pbc.gto.eval_gto import _get_intor_and_comp
from pyscf.pbc.gto.eval_gto import eval_gto as pyscf_eval_gto
from pyscfad.lib import numpy as np
from pyscfad.lib import stop_grad, custom_jvp, vmap
from pyscfad.gto.eval_gto import _eval_gto_fill_grad_r0

def eval_gto(cell, eval_name, coords, comp=None, kpts=None, kpt=None,
             shls_slice=None, non0tab=None, ao_loc=None, out=None):
    if cell.abc is None:
        fn = eval_gto_diff_cell
    else:
        fn = eval_gto_diff_full
    return fn(cell, eval_name, coords, comp=comp, kpts=kpts, kpt=kpt,
              shls_slice=shls_slice, non0tab=non0tab, ao_loc=ao_loc, out=out)

def eval_gto_diff_full(cell, eval_name, coords, comp=None, kpts=None, kpt=None,
                       shls_slice=None, non0tab=None, ao_loc=None, out=None):
    from pyscfad.gto import mole
    from pyscfad.pbc.gto.cell import shift_bas_center
    if eval_name[:3] == 'PBC':  # PBCGTOval_xxx
        eval_name_mol, comp = _get_intor_and_comp(cell, eval_name[3:], comp)
    else:
        eval_name_mol, comp = _get_intor_and_comp(cell, eval_name, comp)

    if kpts is None:
        if kpt is not None:
            kpts_lst = np.reshape(kpt, (1,3))
        else:
            kpts_lst = np.zeros((1,3))
    else:
        kpts_lst = np.reshape(kpts, (-1,3))

    Ls = cell.get_lattice_Ls()
    expkL = np.exp(1j*np.dot(kpts_lst, Ls.T))

    def body(L):
        shifted_cell = shift_bas_center(cell, L)
        ao = mole.eval_gto(shifted_cell, eval_name_mol, coords, comp=comp,
                           shls_slice=shls_slice, non0tab=non0tab,
                           ao_loc=ao_loc, out=None)
        return ao

    # use for loop to reduce memory in fwd-mode AD
    #nk = len(kpts_lst)
    #ng = len(coords)
    #nao = cell.nao
    #if out is None:
    #    if comp == 1:
    #        out = np.zeros((nk,ng,nao), dtype=np.complex128)
    #    else:
    #        out = np.zeros((nk,comp,ng,nao), dtype=np.complex128)
    #for i, L in enumerate(Ls):
    #    ao = body(L)
    #    if comp == 1:
    #        out += expkL[:,i][:,None,None] * ao[None,...]
    #    else:
    #        out += expkL[:,i][:,None,None,None] * ao[None,...]

    aos = np.asarray([body(L) for L in Ls])
    if comp == 1:
        out = np.einsum('kl,lgi->kgi', expkL, aos)
    else:
        out = np.einsum('kl,lcgi->kcgi', expkL, aos)

    if kpts is None or np.shape(kpts) == (3,):  # A single k-point
        out = out[0]
    return out

def eval_gto_diff_cell(cell, eval_name, coords, comp=None, kpts=None, kpt=None,
                       shls_slice=None, non0tab=None, ao_loc=None, out=None):
    if "ip" in eval_name:
        return pyscf_eval_gto(cell, eval_name, coords, comp, kpts, kpt,
                              shls_slice, non0tab, ao_loc, out)
    return _eval_gto(cell, eval_name, coords, comp, kpts, kpt,
                     shls_slice, non0tab, ao_loc, out)

@partial(custom_jvp, nondiff_argnums=tuple(range(1,10)))
def _eval_gto(cell, eval_name, coords, comp, kpts, kpt,
              shls_slice, non0tab, ao_loc, out):
    return pyscf_eval_gto(cell, eval_name, coords, comp=comp, kpts=kpts, kpt=kpt,
                          shls_slice=shls_slice, non0tab=non0tab, ao_loc=ao_loc, out=out)

@_eval_gto.defjvp
def _eval_gto_jvp(eval_name, coords, comp, kpts, kpt,
                  shls_slice, non0tab, ao_loc, out, 
                  primals, tangents):
    cell, = primals
    cell_t, = tangents

    primal_out = _eval_gto(cell, eval_name, coords, comp, kpts, kpt,
                           shls_slice, non0tab, ao_loc, out)

    tangent_out = None
    if cell.coords is not None:
        tangent_out = _eval_gto_jvp_r0(cell, cell_t, eval_name, coords, 
                                       comp, kpts, kpt, shls_slice, non0tab, ao_loc)
    if cell.ctr_coeff is not None:
        pass
    if cell.exp is not None:
        pass
    return primal_out, tangent_out


def _eval_gto_jvp_r0(cell, cell_t, eval_name, coords,
                     comp, kpts, kpt, shls_slice, non0tab, ao_loc):
    if "ip" in eval_name:
        raise NotImplementedError("Please use GTOval_cart_deriv1 or \
                                  GTOval_sph_deriv1 instead of GTOval_ip")
    elif "deriv" in eval_name:
        tmp = eval_name.split("deriv", 1)
        order = int(tmp[1])
        intor_ip = tmp[0] + "deriv" + str(order + 1)
    else:
        #intor_ip = eval_name.replace("GTOval", "GTOval_ip")
        if not "cart" in eval_name or not "sph" in eval_name:
            if cell.cart:
                eval_name = eval_name + "_cart"
            else:
                eval_name = eval_name + "_sph"
        intor_ip = eval_name + "_deriv1"
        order = 0

    ao1 = pyscf_eval_gto(cell, intor_ip, coords, None, kpts, kpt,
                         shls_slice, non0tab, ao_loc, out=None)

    single_kpt = False
    if isinstance(ao1, numpy.ndarray):
        ao1 = [ao1,]
        single_kpt = True
    nkpts = len(ao1)

    if not 'PBC' in intor_ip:
        intor_ip = 'PBC' + intor_ip
    ngrids = len(coords)

    tangent_out = []
    for k in range(nkpts):
        grad = _eval_gto_fill_grad_r0(cell, intor_ip, shls_slice, ao_loc, ao1[k], order, ngrids)
        tangent_out_k = np.einsum("nxlgi,nx->lgi", grad, cell_t.coords)
        grad = None
        if order == 0:
            tangent_out_k = tangent_out_k[0]
        tangent_out.append(tangent_out_k)
    if single_kpt:
        tangent_out = tangent_out[0]
    return tangent_out



from pyscf.df import addons as pyscf_addons
from pyscfad.gto._mole_helper import setup_exp, setup_ctr_coeff

def make_auxmol(mol, auxbasis=None):
    auxmol = pyscf_addons.make_auxmol(mol, auxbasis=auxbasis)
    if mol.exp is not None:
        auxmol.exp, _, _ = setup_exp(auxmol)
    if mol.ctr_coeff is not None:
        auxmol.ctr_coeff, _, _ = setup_ctr_coeff(auxmol)
    return auxmol



from functools import partial
import numpy
from jax import custom_jvp
from pyscf.gto.moleintor import make_loc
from pyscf.gto.eval_gto import _get_intor_and_comp
from pyscf.gto.eval_gto import eval_gto as pyscf_eval_gto
from pyscfad.lib import numpy as np
from pyscfad.lib import ops
from .moleintor import get_bas_label

_MAX_DERIV_ORDER = 4
_DERIV_LABEL = []
for i in range(_MAX_DERIV_ORDER+1):
    if i == 0:
        _label = ["",]
    else:
        _label = get_bas_label(i)
    _DERIV_LABEL += _label

def eval_gto(mol, eval_name, grid_coords,
             comp=None, shls_slice=None, non0tab=None, ao_loc=None, out=None):
    eval_name, comp = _get_intor_and_comp(mol, eval_name, comp)
    return _eval_gto(mol, eval_name, grid_coords,
                     comp, shls_slice, non0tab, ao_loc, out)

@partial(custom_jvp, nondiff_argnums=(1,3,4,5,6,7))
def _eval_gto(mol, eval_name, grid_coords,
              comp, shls_slice, non0tab, ao_loc, out):
    return pyscf_eval_gto(mol, eval_name, grid_coords, comp, shls_slice, non0tab,
                          ao_loc, out)

@_eval_gto.defjvp
def _eval_gto_jvp(eval_name, comp, shls_slice, non0tab, ao_loc, out,
                  primals, tangents):
    mol, grid_coords = primals
    mol_t, grid_coords_t = tangents
    #mol, = primals
    #mol_t, = tangents

    primal_out = _eval_gto(mol, eval_name, grid_coords, comp, shls_slice, non0tab, ao_loc, out)
    tangent_out = np.zeros_like(primal_out)
    nao = primal_out.shape[-1]

    if mol.coords is not None:
        tangent_out += _eval_gto_jvp_r0(mol, mol_t, eval_name, grid_coords,
                                        comp, shls_slice, non0tab, ao_loc)
    if mol.ctr_coeff is not None:
        tangent_out += _eval_gto_jvp_cs(mol, mol_t, eval_name, grid_coords,
                                        comp, shls_slice, non0tab, ao_loc)
    if mol.exp is not None:
        tangent_out += _eval_gto_jvp_exp(mol, mol_t, eval_name, grid_coords,
                                         comp, shls_slice, non0tab, ao_loc)

    tangent_out += _eval_gto_jvp_r(mol, eval_name, grid_coords, grid_coords_t,
                                   comp, shls_slice, non0tab, ao_loc, nao)
    return primal_out, tangent_out

def _eval_gto_jvp_r(mol, eval_name, grid_coords, grid_coords_t,
                    comp, shls_slice, non0tab, ao_loc, nao):
    if "deriv"+str(_MAX_DERIV_ORDER) in eval_name:
        raise NotImplementedError
    if "deriv" not in eval_name:
        new_eval = eval_name + "_deriv1"
        order = 0
    else:
        tmp = eval_name.split("deriv", 1)
        order = int(tmp[1])
        new_eval = tmp[0] + "deriv" + str(order + 1)

    ng = grid_coords.shape[0]
    ao1 = _eval_gto(mol, new_eval, grid_coords, None, shls_slice, non0tab, ao_loc, None)

    nc = (order+1) * (order+2) * (order+3) // 6
    grad = np.zeros((3,nc,ng,nao))
    for iorder in range(order+1):
        start0 = iorder * (iorder+1) * (iorder+2) // 6
        start = (iorder+1) * (iorder+2) * (iorder+3) // 6
        end = (iorder+2) * (iorder+3) * (iorder+4) // 6
        for il, label in enumerate(get_bas_label(iorder)):
            idx_x = _DERIV_LABEL.index("".join(sorted(label + 'x')), start, end)
            idx_y = _DERIV_LABEL.index("".join(sorted(label + 'y')), start, end)
            idx_z = _DERIV_LABEL.index("".join(sorted(label + 'z')), start, end)
            grad = grad.at[0,start0+il].add(ao1[idx_x])
            grad = grad.at[1,start0+il].add(ao1[idx_y])
            grad = grad.at[2,start0+il].add(ao1[idx_z])

    tangent_out = np.einsum("xygi,gx->ygi", grad, grid_coords_t)
    if order == 0:
        tangent_out = tangent_out[0]
    return tangent_out

def _eval_gto_fill_grad_r0(mol, intor, shls_slice, ao_loc, ao1, order, ngrids):
    nc = (order+1) * (order+2) * (order+3) // 6
    if shls_slice is None:
        shls_slice = (0, mol.nbas)
    sh0, sh1 = shls_slice
    if ao_loc is None:
        ao_loc = make_loc(mol._bas, intor)
    ao_start = ao_loc[sh0]
    ao_end = ao_loc[sh1]
    nao = ao_end - ao_start
    ng = ngrids
    atmlst = np.asarray(range(mol.natm))
    aoslices = mol.aoslice_by_atom(ao_loc)

    #if nc == 1:
    #    tangent_out = np.zeros((ng,nao))
    #else:
    #    tangent_out = np.zeros((nc,ng,nao))
    #for iorder in range(order+1):
    #    for k, ia in enumerate(atmlst):
    #        p0, p1 = aoslices [ia, 2:]
    #        if p1 <= ao_start:
    #            continue
    #        id0 = max(0, p0 - ao_start)
    #        id1 = min(p1, ao_end) - ao_start
    #        if order == 0:
    #            tmp = np.einsum('xgi,x->gi', ao1[1:4,:,id0:id1], coords_t[k])
    #            tangent_out = ops.index_add(tangent_out, ops.index[:,id0:id1], tmp)
    #        else:
    #            start0 = iorder * (iorder+1) * (iorder+2) // 6
    #            start = (iorder+1) * (iorder+2) * (iorder+3) // 6
    #            end = (iorder+2) * (iorder+3) * (iorder+4) // 6
    #            for il, label in enumerate(get_bas_label(iorder)):
    #                idx_x = _DERIV_LABEL.index("".join(sorted(label + 'x')), start, end)
    #                idx_y = _DERIV_LABEL.index("".join(sorted(label + 'y')), start, end)
    #                idx_z = _DERIV_LABEL.index("".join(sorted(label + 'z')), start, end)
    #                tmp = (  ao1[idx_x,:,id0:id1] * coords_t[k,0]
    #                       + ao1[idx_y,:,id0:id1] * coords_t[k,1]
    #                       + ao1[idx_z,:,id0:id1] * coords_t[k,2])
    #                tangent_out = ops.index_add(tangent_out, ops.index[start0+il,:,id0:id1], tmp)

    grad = np.zeros([mol.natm,3,nc,ng,nao], dtype=ao1.dtype)
    for iorder in range(order+1):
        for k, ia in enumerate(atmlst):
            p0, p1 = aoslices [ia, 2:]
            if p1 <= ao_start:
                continue
            id0 = max(0, p0 - ao_start)
            id1 = min(p1, ao_end) - ao_start

            start0 = iorder * (iorder+1) * (iorder+2) // 6
            start = (iorder+1) * (iorder+2) * (iorder+3) // 6
            end = (iorder+2) * (iorder+3) * (iorder+4) // 6
            for il, label in enumerate(get_bas_label(iorder)):
                idx_x = _DERIV_LABEL.index("".join(sorted(label + 'x')), start, end)
                idx_y = _DERIV_LABEL.index("".join(sorted(label + 'y')), start, end)
                idx_z = _DERIV_LABEL.index("".join(sorted(label + 'z')), start, end)
                grad = grad.at[k,0,start0+il,:,id0:id1].add(-ao1[idx_x,:,id0:id1])
                grad = grad.at[k,1,start0+il,:,id0:id1].add(-ao1[idx_y,:,id0:id1])
                grad = grad.at[k,2,start0+il,:,id0:id1].add(-ao1[idx_z,:,id0:id1])
    return grad

def _eval_gto_jvp_r0(mol, mol_t, eval_name, grid_coords, comp, shls_slice, non0tab, ao_loc):
    if "deriv"+str(_MAX_DERIV_ORDER) in eval_name:
        raise NotImplementedError
    if "deriv" not in eval_name:
        new_eval = eval_name + "_deriv1"
        order = 0
    else:
        tmp = eval_name.split("deriv", 1)
        order = int(tmp[1])
        new_eval = tmp[0] + "deriv" + str(order + 1)

    ao1 = _eval_gto(mol, new_eval, grid_coords, None, shls_slice, non0tab, ao_loc, None)
    ngrids = len(grid_coords)
    grad = _eval_gto_fill_grad_r0(mol, new_eval, shls_slice, ao_loc, ao1, order, ngrids)
    ao1 = None
    tangent_out = np.einsum("nxlgi,nx->lgi", grad, mol_t.coords)
    grad = None
    if order == 0:
        tangent_out = tangent_out[0]
    return tangent_out

def _eval_gto_jvp_cs(mol, mol_t, eval_name, grid_coords, comp, shls_slice, non0tab, ao_loc):
    return 0

def _eval_gto_jvp_exp(mol, mol_t, eval_name, grid_coords, comp, shls_slice, non0tab, ao_loc):
    return 0



import numpy
from pyscf import __config__
from pyscf import lib as pyscf_lib
from pyscf.pbc.lib.kpts_helper import gamma_point
from pyscf.pbc.df import fft as pyscf_fft
from pyscfad import lib
from pyscfad import util
from pyscfad.lib import numpy as np
from pyscfad.lib import ops
from pyscfad.pbc import tools 
from pyscfad.pbc.gto import Cell

def get_pp(mydf, kpts=None, cell=None):
    '''Get the periodic pseudotential nuc-el AO matrix, with G=0 removed.
    '''
    from pyscf import gto
    from pyscf.pbc.gto import pseudo
    from pyscfad.gto.mole import Mole
    if cell is None:
        cell = mydf.cell
    if kpts is None:
        kpts_lst = numpy.zeros((1,3))
    else:
        kpts_lst = np.reshape(kpts, (-1,3))

    mesh = mydf.mesh
    SI = cell.get_SI()
    Gv = cell.get_Gv(mesh)
    vpplocG = pseudo.get_vlocG(cell, Gv)
    vpplocG = -np.einsum('ij,ij->j', SI, vpplocG)
    ngrids = len(vpplocG)

    # vpploc evaluated in real-space
    vpplocR = tools.ifft(vpplocG, mesh).real
    vpp = [0] * len(kpts_lst)
    for ao_ks_etc, p0, p1 in mydf.aoR_loop(mydf.grids, kpts_lst, cell=cell):
        ao_ks = ao_ks_etc[0]
        for k, ao in enumerate(ao_ks):
            vpp[k] += np.dot(ao.T.conj()*vpplocR[p0:p1], ao)
        ao = ao_ks = None

    # vppnonloc evaluated in reciprocal space
    fakemol = Mole()
    fakemol._atm = numpy.zeros((1,gto.ATM_SLOTS), dtype=numpy.int32)
    fakemol._bas = numpy.zeros((1,gto.BAS_SLOTS), dtype=numpy.int32)
    ptr = gto.PTR_ENV_START
    fakemol._env = numpy.zeros(ptr+10)
    fakemol._bas[0,gto.NPRIM_OF ] = 1
    fakemol._bas[0,gto.NCTR_OF  ] = 1
    fakemol._bas[0,gto.PTR_EXP  ] = ptr+3
    fakemol._bas[0,gto.PTR_COEFF] = ptr+4

    # buf for SPG_lmi upto l=0..3 and nl=3
    #buf = np.empty((48,ngrids), dtype=np.complex128)
    def vppnl_by_k(kpt):
        Gk = Gv + kpt
        #G_rad = np.linalg.norm(Gk, axis=1)
        #G_rad = np.where(G_rad>1e-16, G_rad, 0.)
        absG2 = np.einsum('gx,gx->g', Gk, Gk)
        G_rad = np.where(absG2>1e-16, np.sqrt(np.where(absG2>1e-16, absG2, 0.)), 0.)
        #aokG = ft_ao.ft_ao(cell, Gv, kpt=kpt) * (1/cell.vol)**.5
        # use numerical fft for now
        coords = mydf.grids.coords
        aoR = cell.pbc_eval_gto('PBCGTOval', coords, kpt=kpt)
        assert numpy.prod(mesh) == len(coords) == ngrids
        aokG = tools.fftk(aoR.T, mesh, np.exp(-1j*np.dot(coords, kpt))).T

        vppnl = 0
        for ia in range(cell.natm):
            symb = cell.atom_symbol(ia)
            if symb not in cell._pseudo:
                continue
            pp = cell._pseudo[symb]
            p1 = 0
            buf = []
            for l, proj in enumerate(pp[5:]):
                rl, nl, hl = proj
                if nl > 0:
                    fakemol._bas[0,gto.ANG_OF] = l
                    fakemol._env[ptr+3] = .5*rl**2
                    fakemol._env[ptr+4] = rl**(l+1.5)*numpy.pi**1.25
                    pYlm_part = fakemol.eval_gto('GTOval', Gk)

                    p0, p1 = p1, p1+nl*(l*2+1)
                    # pYlm is real, SI[ia] is complex
                    #pYlm = np.array((nl,l*2+1,ngrids), dtype=numpy.complex128, buffer=buf[p0:p1])
                    pYlm = []
                    for k in range(nl):
                        qkl = pseudo.pp._qli(G_rad*rl, l, k)
                        pYlm.append(pYlm_part.T * qkl)
                    buf.append(np.asarray(pYlm).reshape(-1,ngrids))

                    #:SPG_lmi = numpy.einsum('g,nmg->nmg', SI[ia].conj(), pYlm)
                    #:SPG_lm_aoG = numpy.einsum('nmg,gp->nmp', SPG_lmi, aokG)
                    #:tmp = numpy.einsum('ij,jmp->imp', hl, SPG_lm_aoG)
                    #:vppnl += numpy.einsum('imp,imq->pq', SPG_lm_aoG.conj(), tmp)

            buf = np.vstack(buf)
            if p1 > 0:
                #SPG_lmi = buf #buf[:p1]
                SPG_lmi = buf * SI[ia].conj()
                SPG_lm_aoGs = np.dot(SPG_lmi, aokG)
                p1 = 0
                for l, proj in enumerate(pp[5:]):
                    rl, nl, hl = proj
                    if nl > 0:
                        p0, p1 = p1, p1+nl*(l*2+1)
                        hl = numpy.asarray(hl)
                        SPG_lm_aoG = SPG_lm_aoGs[p0:p1].reshape(nl,l*2+1,-1)
                        tmp = np.einsum('ij,jmp->imp', hl, SPG_lm_aoG)
                        vppnl += np.einsum('imp,imq->pq', SPG_lm_aoG.conj(), tmp)
        #return vppnl * (1./cell.vol)
        return vppnl * (1./ngrids**2)

    for k, kpt in enumerate(kpts_lst):
        vppnl = vppnl_by_k(kpt)
        if gamma_point(kpt):
            vpp[k] = vpp[k].real + vppnl.real
        else:
            vpp[k] += vppnl

    if kpts is None or numpy.shape(kpts) == (3,):
        vpp = vpp[0]
    return np.asarray(vpp)


@util.pytree_node(['cell','kpts'])
class FFTDF(pyscf_fft.FFTDF):
    def __init__(self, cell, kpts=numpy.zeros((1,3)), **kwargs):
        from pyscf.pbc.dft import gen_grid
        from pyscfad.pbc.dft import numint
        self.cell = cell
        self.stdout = cell.stdout
        self.verbose = cell.verbose
        self.max_memory = cell.max_memory

        self.kpts = kpts
        self.grids = gen_grid.UniformGrids(cell)

        # to mimic molecular DF object
        self.blockdim = getattr(__config__, 'pbc_df_df_DF_blockdim', 240)

        # The following attributes are not input options.
        # self.exxdiv has no effects. It was set in the get_k_kpts function to
        # mimic the KRHF/KUHF object in the call to tools.get_coulG.
        self.exxdiv = None
        self._numint = numint.KNumInt()
        self._rsh_df = {}  # Range separated Coulomb DF objects
        self._keys = set(self.__dict__.keys())
        self.__dict__.update(kwargs)

    def get_jk(self, dm, hermi=1, kpts=None, kpts_band=None,
               with_j=True, with_k=True, omega=None, exxdiv=None, cell=None):
        from pyscfad.pbc.df import fft_jk
        if omega is not None:  # J/K for RSH functionals
            raise NotImplementedError
            #return _sub_df_jk_(self, dm, hermi, kpts, kpts_band,
            #                   with_j, with_k, omega, exxdiv)

        if kpts is None:
            if numpy.all(self.kpts == 0): # Gamma-point J/K by default
                kpts = numpy.zeros(3)
            else:
                kpts = self.kpts
        else:
            kpts = np.asarray(kpts)

        vj = vk = None
        if kpts.shape == (3,):
            vj, vk = fft_jk.get_jk(self, dm, hermi, kpts, kpts_band,
                                   with_j, with_k, exxdiv, cell=cell)
        else:
            if with_k:
                vk = fft_jk.get_k_kpts(self, dm, hermi, kpts, kpts_band, exxdiv, cell=cell)
            if with_j:
                vj = fft_jk.get_j_kpts(self, dm, hermi, kpts, kpts_band, cell=cell)
        return vj, vk

    get_pp = get_pp



import warnings
from jax import tree_util
from pyscf import __config__

PYSCFAD = getattr(__config__, "pyscfad", False)

def pytree_node(leaf_names, num_args=0):
    '''
    Class decorator that register the underlying class as a pytree node.
    See `jax document<https://jax.readthedocs.io/en/latest/pytrees.html>`_
    for the definition of pytrees.

    Args:
        leaf_names : list or tuple
            Attributes of the class that are traced as pytree leaves.

    Kwargs:
        num_args: int
            Number of positional arguments in ``leaf_names``.
            This is useful when the ``__init__`` method of the class
            has positional arguments that are named differently than
            the actual attribute names. Default value is 0.

    Note:
        The ``__init__`` method of the class can't have positional arguments
        that are not included in ``leaf_names``. If ``num_args`` is greater
        than 0, the sequence of positional arguments in ``leaf_names`` must
        follow that in the ``__init__`` method.
    '''
    def class_orig(cls):
        return cls

    def class_as_pytree_node(cls):
        def tree_flatten(obj):
            keys = obj.__dict__.keys()
            for leaf_name in leaf_names:
                if leaf_name not in keys:
                    raise KeyError(f"Pytree leaf {leaf_name} is not defined in class {cls}.")
            children =  tuple(getattr(obj, leaf_name, None) for leaf_name in leaf_names)
            if len(children) <= 0:
                #raise KeyError("Empty pytree node is not supported.")
                warnings.warn(f"Not taking derivatives wrt the leaves in "
                              f"the node {obj.__class__} as none of those was specified.")

            aux_keys = set(keys) - set(leaf_names)
            aux_data = tuple(getattr(obj, key, None) for key in aux_keys)
            metadata = (num_args,) + tuple(zip(aux_keys, aux_data))
            return children, metadata

        def tree_unflatten(metadata, children):
            num_args = metadata[0]
            metadata = metadata[1:]
            leaves_args = children[:num_args]
            leaves_kwargs = tuple(zip(leaf_names[num_args:], children[num_args:]))
            kwargs = dict(leaves_kwargs + metadata)
            obj = cls(*leaves_args, **kwargs)
            return obj

        tree_util.register_pytree_node(cls, tree_flatten, tree_unflatten)
        return cls

    if PYSCFAD:
        return class_as_pytree_node
    else:
        return class_orig



import tempfile
from pyscf import __config__
from pyscf import lib as pyscf_lib
from pyscf.lib import logger
from pyscf.df import df as pyscf_df
from pyscfad import lib
from pyscfad import util
from pyscfad.df import addons, incore, df_jk

@util.pytree_node(['mol', 'auxmol', '_cderi'], num_args=1)
class DF(pyscf_df.DF):
    def __init__(self, mol, auxbasis=None, **kwargs):
        pyscf_df.DF.__init__(self, mol, auxbasis=auxbasis)
        self.__dict__.update(kwargs)

    def build(self):
        log = logger.new_logger(self)

        self.check_sanity()
        self.dump_flags()

        mol = self.mol
        if self.auxmol is None:
            self.auxmol = addons.make_auxmol(self.mol, self.auxbasis)
        auxmol = self.auxmol
        nao = mol.nao
        naux = auxmol.nao
        nao_pair = nao*nao

        max_memory = self.max_memory - pyscf_lib.current_memory()[0]
        int3c = mol._add_suffix('int3c2e')
        int2c = mol._add_suffix('int2c2e')
        if (nao_pair*naux*8/1e6 < .9*max_memory and
            not isinstance(self._cderi_to_save, str)):
            self._cderi = incore.cholesky_eri(mol, auxmol=auxmol,
                                              int3c=int3c, int2c=int2c,
                                              max_memory=max_memory, verbose=log)
        else:
            raise NotImplementedError("Outcore density fitting is not implemented.")
        del log
        return self

    def reset(self, mol=None):
        '''Reset mol and clean up relevant attributes for scanner mode'''
        if mol is not None:
            self.mol = mol
        # resetting auxmol will lose its tracing
        #self.auxmol = None
        self._cderi = None
        if not isinstance(self._cderi_to_save, str):
            # pylint: disable=consider-using-with
            self._cderi_to_save = tempfile.NamedTemporaryFile(dir=pyscf_lib.param.TMPDIR)
        self._vjopt = None
        self._rsh_df = {}
        return self

    def get_naoaux(self):
        # determine naoaux with self._cderi, because DF object may be used as CD
        # object when self._cderi is provided.
        if self._cderi is None:
            self.build()
        return self._cderi.shape[0]

    def get_jk(self, dm, hermi=1, with_j=True, with_k=True,
               direct_scf_tol=getattr(__config__, 'scf_hf_SCF_direct_scf_tol', 1e-13),
               omega=None):
        if self._cderi is None:
            self.build()
        if omega is None:
            return df_jk.get_jk(self, dm, hermi, with_j, with_k, direct_scf_tol)
        else:
            raise NotImplementedError



import numpy
from jax import vmap, jit
from pyscf import lib as pyscf_lib
from pyscf.lib import logger
from pyscf import df as pyscf_df
from pyscf.gw import rpa as pyscf_rpa
from pyscfad import util
from pyscfad.lib import numpy as np
from pyscfad import gto, scf, dft, df


def kernel(rpa, mo_energy, mo_coeff, Lpq=None, nw=None, verbose=logger.NOTE):
    mf = rpa._scf
    # only support frozen core
    if rpa.frozen is not None:
        assert isinstance(rpa.frozen, int)
        assert rpa.frozen < rpa.nocc

    if Lpq is None:
        Lpq = rpa.ao2mo(mo_coeff)

    # Grids for integration on imaginary axis
    freqs, wts = pyscf_rpa._get_scaled_legendre_roots(nw)

    # Compute HF exchange energy (EXX)
    dm = mf.make_rdm1()
    rhf = scf.RHF(rpa.mol)
    e_hf = rhf.energy_elec(dm=dm)[0]
    e_hf += mf.energy_nuc()

    # Compute RPA correlation energy
    e_corr = get_rpa_ecorr(rpa, Lpq, freqs, wts)

    # Compute totol energy
    e_tot = e_hf + e_corr

    logger.debug(rpa, '  RPA total energy = %s', e_tot)
    logger.debug(rpa, '  EXX energy = %s, RPA corr energy = %s', e_hf, e_corr)

    return e_tot, e_hf, e_corr

@jit
def get_rho_response(omega, mo_energy, Lpq):
    """
    Compute density response function in auxiliary basis at freq iw.
    """
    nocc = Lpq.shape[1]
    eia = mo_energy[:nocc, None] - mo_energy[None, nocc:]
    eia = eia / (omega**2 + eia * eia)
    # Response from both spin-up and spin-down density
    Pia = Lpq * (eia * 4.0)
    Pi = np.einsum('Pia, Qia -> PQ', Pia, Lpq)
    return Pi

def get_rpa_ecorr(rpa, Lpq, freqs, wts):
    """
    Compute RPA correlation energy
    """
    mol = rpa.mol
    mf = rpa._scf
    dm = mf.make_rdm1()
    rks = dft.RKS(mol, xc=mf.xc)
    veff = rks.get_veff(mol, dm)
    h1e = rks.get_hcore(mol)
    s1e = rks.get_ovlp(mol)
    fock = rks.get_fock(h1e, s1e, veff, dm)
    mo_energy, _ = rks.eig(fock, s1e)

    #mo_energy = _mo_energy_without_core(rpa, rpa._scf.mo_energy)
    mo_energy = pyscf_rpa._mo_energy_without_core(rpa, mo_energy)
    nocc = rpa.nocc
    naux = Lpq.shape[0]

    if (mo_energy[nocc] - mo_energy[nocc-1]) < 1e-3:
        logger.warn(rpa, 'Current RPA code not well-defined for degeneracy!')

    def body(omega, weight):
        Pi = get_rho_response(omega, mo_energy, Lpq[:, :nocc, nocc:])
        ec_w  = np.log(np.linalg.det(np.eye(naux) - Pi))
        ec_w += np.trace(Pi)
        e_corr_i = 1./(2.*numpy.pi) * ec_w * weight
        return e_corr_i
    e_corr_i = vmap(body)(freqs, wts)
    e_corr = np.sum(e_corr_i)
    return e_corr

@util.pytree_node(['_scf','mol','with_df','mo_energy','mo_coeff'], num_args=1)
class RPA(pyscf_rpa.RPA):
    def __init__(self, mf, frozen=None, auxbasis=None, **kwargs):
        self.mol = mf.mol
        self._scf = mf
        self.verbose = self.mol.verbose
        self.stdout = self.mol.stdout
        self.max_memory = mf.max_memory
        self.frozen = frozen
        self.with_df = None

##################################################
# don't modify the following attributes, they are not input options
        self._nocc = None
        self._nmo = None
        self.mo_energy = mf.mo_energy
        self.mo_coeff = mf.mo_coeff
        self.mo_occ = mf.mo_occ
        self.e_corr = None
        self.e_hf = None
        self.e_tot = None

        self.__dict__.update(kwargs)
        if self.with_df is None:
            if getattr(self._scf, 'with_df', None):
                self.with_df = self._scf.with_df
            else:
                if auxbasis is None:
                    auxbasis = pyscf_df.addons.make_auxbasis(self.mol, mp2fit=True)
                auxmol = df.addons.make_auxmol(self.mol, auxbasis)
                self.with_df = df.DF(self.mol, auxmol=auxmol)

    def kernel(self, mo_energy=None, mo_coeff=None, Lpq=None, nw=40):
        """
        Args:
            mo_energy : 1D array (nmo), mean-field mo energy
            mo_coeff : 2D array (nmo, nmo), mean-field mo coefficient
            Lpq : 3D array (naux, nmo, nmo), 3-index ERI
            nw: interger, grid number
        Returns:
            self.e_tot : RPA total eenrgy
            self.e_hf : EXX energy
            self.e_corr : RPA correlation energy
        """
        if mo_coeff is None:
            mo_coeff = pyscf_rpa._mo_without_core(self, self._scf.mo_coeff)
        if mo_energy is None:
            mo_energy = pyscf_rpa._mo_energy_without_core(self, self._scf.mo_energy)

        log = logger.new_logger(self)
        self.dump_flags()
        self.e_tot, self.e_hf, self.e_corr = \
                        kernel(self, mo_energy, mo_coeff, Lpq=Lpq, nw=nw, verbose=self.verbose)

        log.timer('RPA')
        del log
        return self.e_corr

    def ao2mo(self, mo_coeff=None):
        if mo_coeff is None:
            mo_coeff = self.mo_coeff
        nmo = self.nmo
        naux = self.with_df.get_naoaux()
        mem_incore = (2 * nmo**2*naux) * 8 / 1e6
        mem_now = pyscf_lib.current_memory()[0]

        if (mem_incore + mem_now < 0.99 * self.max_memory) or self.mol.incore_anyway:
            Lpq = np.einsum("lpq,pi,qj->lij", self.with_df._cderi, mo_coeff, mo_coeff)
            return Lpq
        else:
            raise RuntimeError("not enough memory")



import numpy
from pyscf.pbc import tools as pyscf_tools
from pyscf.pbc.lib.kpts_helper import is_zero, member
from pyscfad.lib import numpy as np
from pyscfad.lib import ops

def _ewald_exxdiv_for_G0(cell, kpts, dms, vk, kpts_band=None):
    s = cell.pbc_intor('int1e_ovlp', hermi=1, kpts=kpts)
    madelung = pyscf_tools.pbc.madelung(cell, kpts)
    if kpts is None:
        for i,dm in enumerate(dms):
            #vk[i] += madelung * reduce(numpy.dot, (s, dm, s))
            vk = ops.index_add(vk, ops.index[i],
                               madelung * np.dot(s, np.dot(dm, s)))
    elif numpy.shape(kpts) == (3,):
        if kpts_band is None or is_zero(kpts_band-kpts):
            for i,dm in enumerate(dms):
                #vk[i] += madelung * reduce(numpy.dot, (s, dm, s))
                vk = ops.index_add(vk, ops.index[i],
                                   madelung * np.dot(s, np.dot(dm, s)))
    elif kpts_band is None or numpy.array_equal(kpts, kpts_band):
        for k in range(len(kpts)):
            for i,dm in enumerate(dms):
                #vk[i,k] += madelung * reduce(numpy.dot, (s[k], dm[k], s[k]))
                vk = ops.index_add(vk, ops.index[i,k],
                                   madelung * np.dot(s[k], np.dot(dm[k], s[k])))
    else:
        for k, kpt in enumerate(kpts):
            for kp in member(kpt, kpts_band.reshape(-1,3)):
                for i,dm in enumerate(dms):
                    #vk[i,kp] += madelung * reduce(numpy.dot, (s[k], dm[k], s[k]))
                    vk = ops.index_add(vk, ops.index[i,kp],
                                   madelung * np.dot(s[k], np.dot(dm[k], s[k])))
    return vk



from pyscf import numpy as np
from pyscfad.lib import jit

def full(eri_ao, mo_coeff, verbose=0, compact=True, **kwargs):
    nao = mo_coeff.shape[0]
    if eri_ao.size != nao**4:
        raise NotImplementedError
    return general(eri_ao, (mo_coeff,)*4, verbose, compact)

@jit
def general(eri_ao, mo_coeffs, verbose=0, compact=True, **kwargs):
    nao = mo_coeffs[0].shape[0]
    #return np.einsum('pqrs,pi,qj,rk,sl->ijkl', eri_ao.reshape([nao]*4),
    #                  mo_coeffs[0].conj(), mo_coeffs[1],
    #                  mo_coeffs[2].conj(), mo_coeffs[3])
    eri_pqrl = np.dot(eri_ao.reshape(-1,nao), mo_coeffs[3])
    eri_lpqk = np.dot(eri_pqrl.T.reshape(-1,nao), mo_coeffs[2].conj())
    eri_klpj = np.dot(eri_lpqk.T.reshape(-1,nao), mo_coeffs[1])
    eri_jkli = np.dot(eri_klpj.T.reshape(-1,nao), mo_coeffs[0].conj())
    eri_ijkl = eri_jkli.T.reshape([nao]*4)
    return eri_ijkl



from pyscf import numpy as np
from pyscf.lib import logger
from pyscf.lib import current_memory
from pyscfad import util
from pyscfad.lib import jit
from pyscfad import ao2mo
from pyscfad.cc import ccsd
from pyscfad.cc import rintermediates as imd

@jit
def update_amps(cc, t1, t2, eris):
    nocc, nvir = t1.shape
    fock = eris.fock
    mo_e_o = eris.mo_energy[:nocc]
    mo_e_v = eris.mo_energy[nocc:] + cc.level_shift
    mo_oo = np.diagflat(mo_e_o)
    mo_vv = np.diagflat(mo_e_v)

    foo = fock[:nocc,:nocc]#.copy()
    fvv = fock[nocc:,nocc:]#.copy()

    t1new, t2new = cc.amplitude_equation(t1, t2, eris)
    # Move energy terms to the other side
    t1new +=   np.einsum('ac,ic->ia', -mo_vv, t1)
    t1new +=  -np.einsum('ki,ka->ia', -mo_oo, t1)
    if cc.cc2:
        Lvv2 = -np.diagflat(np.diag(fvv))
        tmp = np.einsum('ac,ijcb->ijab', Lvv2, t2)
        t2new += (tmp + tmp.transpose(1,0,3,2))
        Loo2 = -np.diagflat(np.diag(foo))
        tmp = np.einsum('ki,kjab->ijab', Loo2, t2)
        t2new -= (tmp + tmp.transpose(1,0,3,2))
    else:
        tmp = np.einsum('ac,ijcb->ijab', -mo_vv, t2)
        t2new += tmp + tmp.transpose(1,0,3,2)
        tmp = np.einsum('ki,kjab->ijab', -mo_oo, t2)
        t2new -= tmp + tmp.transpose(1,0,3,2)

    eia = mo_e_o[:,None] - mo_e_v
    #eijab = direct_sum('ia,jb->ijab',eia,eia)
    eijab = eia[:,None,:,None] + eia[None,:,None,:]
    t1new /= eia
    t2new /= eijab
    return t1new, t2new

@jit
def amplitude_equation(cc, t1, t2, eris):
    nocc, nvir = t1.shape
    fock = eris.fock

    fov = fock[:nocc,nocc:]#.copy()
    foo = fock[:nocc,:nocc]#.copy()
    fvv = fock[nocc:,nocc:]#.copy()

    Foo = imd.cc_Foo(t1,t2,eris)
    Fvv = imd.cc_Fvv(t1,t2,eris)
    Fov = imd.cc_Fov(t1,t2,eris)

    # T1 equation
    t1new  =-2*np.einsum('kc,ka,ic->ia', fov, t1, t1)
    t1new +=   np.einsum('ac,ic->ia', Fvv, t1)
    t1new +=  -np.einsum('ki,ka->ia', Foo, t1)
    t1new += 2*np.einsum('kc,kica->ia', Fov, t2)
    t1new +=  -np.einsum('kc,ikca->ia', Fov, t2)
    t1new +=   np.einsum('kc,ic,ka->ia', Fov, t1, t1)
    t1new += fov.conj()
    t1new += 2*np.einsum('kcai,kc->ia', eris.ovvo, t1)
    t1new +=  -np.einsum('kiac,kc->ia', eris.oovv, t1)
    eris_ovvv = eris.get_ovvv()
    t1new += 2*np.einsum('kdac,ikcd->ia', eris_ovvv, t2)
    t1new +=  -np.einsum('kcad,ikcd->ia', eris_ovvv, t2)
    t1new += 2*np.einsum('kdac,kd,ic->ia', eris_ovvv, t1, t1)
    t1new +=  -np.einsum('kcad,kd,ic->ia', eris_ovvv, t1, t1)
    eris_ovoo = eris.ovoo
    t1new +=-2*np.einsum('lcki,klac->ia', eris_ovoo, t2)
    t1new +=   np.einsum('kcli,klac->ia', eris_ovoo, t2)
    t1new +=-2*np.einsum('lcki,lc,ka->ia', eris_ovoo, t1, t1)
    t1new +=   np.einsum('kcli,lc,ka->ia', eris_ovoo, t1, t1)

    # T2 equation
    tmp2  = np.einsum('kibc,ka->abic', eris.oovv, -t1)
    tmp2 += np.asarray(eris_ovvv).conj().transpose(1,3,0,2)
    tmp = np.einsum('abic,jc->ijab', tmp2, t1)
    t2new = tmp + tmp.transpose(1,0,3,2)
    tmp2  = np.einsum('kcai,jc->akij', eris.ovvo, t1)
    tmp2 += eris_ovoo.transpose(1,3,0,2).conj()
    tmp = np.einsum('akij,kb->ijab', tmp2, t1)
    t2new -= tmp + tmp.transpose(1,0,3,2)
    t2new += np.asarray(eris.ovov).conj().transpose(0,2,1,3)
    if cc.cc2:
        Woooo2 = eris.oooo.transpose(0,2,1,3)
        Woooo2 += np.einsum('lcki,jc->klij', eris_ovoo, t1)
        Woooo2 += np.einsum('kclj,ic->klij', eris_ovoo, t1)
        Woooo2 += np.einsum('kcld,ic,jd->klij', eris.ovov, t1, t1)
        t2new += np.einsum('klij,ka,lb->ijab', Woooo2, t1, t1)
        Wvvvv = np.einsum('kcbd,ka->abcd', eris_ovvv, -t1)
        Wvvvv = Wvvvv + Wvvvv.transpose(1,0,3,2)
        Wvvvv += eris.vvvv.transpose(0,2,1,3)
        t2new += np.einsum('abcd,ic,jd->ijab', Wvvvv, t1, t1)
        Lvv2 = fvv - np.einsum('kc,ka->ac', fov, t1)
        #Lvv2 -= np.diagflat(np.diag(fvv))
        tmp = np.einsum('ac,ijcb->ijab', Lvv2, t2)
        t2new += (tmp + tmp.transpose(1,0,3,2))
        Loo2 = foo + np.einsum('kc,ic->ki', fov, t1)
        #Loo2 -= np.diagflat(np.diag(foo))
        tmp = np.einsum('ki,kjab->ijab', Loo2, t2)
        t2new -= (tmp + tmp.transpose(1,0,3,2))
    else:
        Loo = imd.Loo(t1, t2, eris)
        Lvv = imd.Lvv(t1, t2, eris)

        Woooo = imd.cc_Woooo(t1, t2, eris)
        Wvoov = imd.cc_Wvoov(t1, t2, eris)
        Wvovo = imd.cc_Wvovo(t1, t2, eris)
        Wvvvv = imd.cc_Wvvvv(t1, t2, eris)

        tau = t2 + np.einsum('ia,jb->ijab', t1, t1)
        t2new += np.einsum('klij,klab->ijab', Woooo, tau)
        t2new += np.einsum('abcd,ijcd->ijab', Wvvvv, tau)
        tmp = np.einsum('ac,ijcb->ijab', Lvv, t2)
        t2new += tmp + tmp.transpose(1,0,3,2)
        tmp = np.einsum('ki,kjab->ijab', Loo, t2)
        t2new -= tmp + tmp.transpose(1,0,3,2)
        tmp  = 2.*np.einsum('akic,kjcb->ijab', Wvoov, t2)
        tmp -=   np.einsum('akci,kjcb->ijab', Wvovo, t2)
        t2new += tmp + tmp.transpose(1,0,3,2)
        tmp = np.einsum('akic,kjbc->ijab', Wvoov, t2)
        t2new -= tmp + tmp.transpose(1,0,3,2)
        tmp = np.einsum('bkci,kjac->ijab', Wvovo, t2)
        t2new -= tmp + tmp.transpose(1,0,3,2)
    return t1new, t2new

@util.pytree_node(ccsd.CC_Tracers, num_args=1)
class RCCSD(ccsd.CCSD):
    def kernel(self, t1=None, t2=None, eris=None, mbpt2=False):
        return self.ccsd(t1, t2, eris, mbpt2)

    def ccsd(self, t1=None, t2=None, eris=None, mbpt2=False):
        if mbpt2:
            raise NotImplementedError

        if eris is None:
            eris = self.ao2mo(self.mo_coeff)
        return ccsd.CCSD.ccsd(self, t1, t2, eris)

    def ao2mo(self, mo_coeff=None):
        nmo = self.nmo
        nao = self.mo_coeff.shape[0]
        nmo_pair = nmo * (nmo+1) // 2
        nao_pair = nao * (nao+1) // 2
        mem_incore = (max(nao_pair**2, nmo**4) + nmo_pair**2) * 8/1e6
        mem_now = current_memory()[0]
        if (self._scf._eri is not None and
            (mem_incore+mem_now < self.max_memory) or self.mol.incore_anyway):
            return _make_eris_incore(self, mo_coeff)
        elif getattr(self._scf, 'with_df', None):
            raise NotImplementedError
        else:
            raise NotImplementedError

    update_amps = update_amps
    amplitude_equation = amplitude_equation

def _make_eris_incore(mycc, mo_coeff=None, ao2mofn=None):
    log = logger.new_logger(mycc)
    eris = _ChemistsERIs()
    eris._common_init_(mycc, mo_coeff)
    nocc = eris.nocc
    nmo = eris.fock.shape[0]

    if callable(ao2mofn):
        eri1 = ao2mofn(eris.mo_coeff).reshape([nmo]*4)
    else:
        eri1 = ao2mo.incore.full(mycc._scf._eri, eris.mo_coeff, compact=False)
        #eri1 = ao2mo.restore(1, eri1, nmo)
    eris.oooo = eri1[:nocc,:nocc,:nocc,:nocc]#.copy()
    eris.ooov = eri1[:nocc,:nocc,:nocc,nocc:]
    eris.ovoo = eri1[:nocc,nocc:,:nocc,:nocc]#.copy()
    eris.ovov = eri1[:nocc,nocc:,:nocc,nocc:]#.copy()
    eris.oovv = eri1[:nocc,:nocc,nocc:,nocc:]#.copy()
    eris.ovvo = eri1[:nocc,nocc:,nocc:,:nocc]#.copy()
    eris.ovvv = eri1[:nocc,nocc:,nocc:,nocc:]#.copy()
    eris.vvvv = eri1[nocc:,nocc:,nocc:,nocc:]#.copy()
    log.timer('CCSD integral transformation')
    del(log)
    return eris

@util.pytree_node(ccsd.ERI_Tracers)
class _ChemistsERIs(ccsd._ChemistsERIs):
    def get_ovvv(self, *slices):
        '''To access a subblock of ovvv tensor'''
        if slices:
            return self.ovvv[slices]
        else:
            return self.ovvv



import numpy
from pyscf import __config__
from pyscf.lib import current_memory
from pyscf.lib import logger
from pyscf.dft import rks as pyscf_rks
from pyscf.dft import gen_grid
from pyscfad import lib
from pyscfad import util
from pyscfad.lib import numpy as np
from pyscfad.lib import stop_grad
from pyscfad.scf import hf
from pyscfad.dft import numint

@util.pytree_node([])
class VXC():
    def __init__(self, **kwargs):
        self.vxc = None
        self.ecoul = None
        self.exc = None
        self.vj = None
        self.vk = None
        self.__dict__.update(kwargs)

    def __repr__(self):
        return self.vxc.__repr__()

def get_veff(ks, mol=None, dm=None, dm_last=0, vhf_last=0, hermi=1):
    if mol is None:
        mol = ks.mol
    if dm is None:
        dm = ks.make_rdm1()
    log = logger.new_logger(ks)

    ground_state = getattr(dm, "ndim", None) == 2

    if ks.grids.coords is None:
        ks.grids.build(with_non0tab=True)
        if ks.small_rho_cutoff > 1e-20 and ground_state:
            # Filter grids the first time setup grids
            ks.grids = prune_small_rho_grids_(ks, mol, dm, ks.grids)
        log.timer('setting up grids')
    if ks.nlc != '':
        if ks.nlcgrids.coords is None:
            ks.nlcgrids.build(with_non0tab=True)
            if ks.small_rho_cutoff > 1e-20 and ground_state:
                # Filter grids the first time setup grids
                ks.nlcgrids = prune_small_rho_grids_(ks, mol, dm, ks.nlcgrids)
            log.timer('setting up nlc grids')

    ni = ks._numint
    vxc = VXC()
    if hermi == 2:  # because rho = 0
        n, vxc.exc, vxc.vxc = 0, 0., 0.
    else:
        max_memory = ks.max_memory - current_memory()[0]
        n, vxc.exc, vxc.vxc = ni.nr_rks(mol, ks.grids, ks.xc, dm, max_memory=max_memory)
        if ks.nlc != '':
            assert 'VV10' in ks.nlc.upper()
            _, enlc, vnlc = ni.nr_rks(mol, ks.nlcgrids, ks.xc+'__'+ks.nlc, dm,
                                      max_memory=max_memory)
            vxc.exc += enlc
            vxc.vxc += vnlc
        log.debug('nelec by numeric integration = %s', n)
        log.timer('vxc')

    #enabling range-separated hybrids
    omega, alpha, hyb = ni.rsh_and_hybrid_coeff(ks.xc, spin=mol.spin)
    if abs(hyb) < 1e-10 and abs(alpha) < 1e-10:
        vk = None
        if (ks._eri is None and ks.direct_scf and
            getattr(vhf_last, 'vj', None) is not None):
            ddm = np.asarray(dm) - np.asarray(dm_last)
            vj = ks.get_j(mol, ddm, hermi)
            vj += vhf_last.vj
        else:
            vj = ks.get_j(mol, dm, hermi)
        vxc.vxc += vj
    else:
        if (ks._eri is None and ks.direct_scf and
            getattr(vhf_last, 'vk', None) is not None):
            ddm = np.asarray(dm) - np.asarray(dm_last)
            vj, vk = ks.get_jk(mol, ddm, hermi)
            vk *= hyb
            if abs(omega) > 1e-10:  # For range separated Coulomb operator
                vklr = ks.get_k(mol, ddm, hermi, omega=omega)
                vklr *= (alpha - hyb)
                vk += vklr
            vj += vhf_last.vj
            vk += vhf_last.vk
        else:
            vj, vk = ks.get_jk(mol, dm, hermi)
            vk *= hyb
            if abs(omega) > 1e-10:
                vklr = ks.get_k(mol, dm, hermi, omega=omega)
                vklr *= (alpha - hyb)
                vk += vklr
        vxc.vxc += vj - vk * .5

        if ground_state:
            vxc.exc -= np.einsum('ij,ji', dm, vk).real * .5 * .5
    if ground_state:
        vxc.ecoul = np.einsum('ij,ji', dm, vj).real * .5
    else:
        vxc.ecoul = None

    del log
    return vxc

def energy_elec(ks, dm=None, h1e=None, vhf=None):
    if dm is None:
        dm = ks.make_rdm1()
    if h1e is None:
        h1e = ks.get_hcore()
    if vhf is None or getattr(vhf, 'ecoul', None) is None:
        vhf = ks.get_veff(ks.mol, dm)
    e1 = np.einsum('ij,ji->', h1e, dm)
    e2 = vhf.ecoul + vhf.exc
    ks.scf_summary['e1'] = stop_grad(e1).real
    ks.scf_summary['coul'] = stop_grad(vhf.ecoul).real
    ks.scf_summary['exc'] = stop_grad(vhf.exc).real
    logger.debug(ks, 'E1 = %s  Ecoul = %s  Exc = %s', e1, vhf.ecoul, vhf.exc)
    return (e1+e2).real, e2

NELEC_ERROR_TOL = getattr(__config__, 'dft_rks_prune_error_tol', 0.02)
def prune_small_rho_grids_(ks, mol, dm, grids):
    mol = stop_grad(mol)
    dm = stop_grad(dm)
    rho = ks._numint.get_rho(mol, dm, grids, ks.max_memory)
    n = numpy.dot(rho, grids.weights)
    if abs(n-mol.nelectron) < NELEC_ERROR_TOL*n:
        rho *= grids.weights
        idx = abs(rho) > ks.small_rho_cutoff / grids.weights.size
        logger.debug(ks, 'Drop grids %d',
                     grids.weights.size - numpy.count_nonzero(idx))
        grids.coords  = numpy.asarray(grids.coords [idx], order='C')
        grids.weights = numpy.asarray(grids.weights[idx], order='C')
        grids.non0tab = grids.make_mask(mol, grids.coords)
    return grids

def _dft_common_init_(mf, xc='LDA,VWN'):
    mf.xc = xc
    mf.nlc = ''
    mf.grids = None
    mf.nlcgrids = None
    # Use rho to filter grids
    mf.small_rho_cutoff = getattr(__config__, 'dft_rks_RKS_small_rho_cutoff', 1e-7)
##################################################
# don't modify the following attributes, they are not input options
    mf._numint = numint.NumInt()
    mf._keys = mf._keys.union(['xc', 'nlc', 'omega', 'grids', 'nlcgrids',
                               'small_rho_cutoff'])

def _dft_common_post_init_(mf):
    if mf.grids is None:
        mf.grids = gen_grid.Grids(stop_grad(mf.mol))
        mf.grids.level = getattr(__config__, 'dft_rks_RKS_grids_level',
                                 mf.grids.level)
    if mf.nlcgrids is None:
        mf.nlcgrids = gen_grid.Grids(stop_grad(mf.mol))
        mf.nlcgrids.level = getattr(__config__, 'dft_rks_RKS_nlcgrids_level',
                                    mf.nlcgrids.level)


class KohnShamDFT(pyscf_rks.KohnShamDFT):
    __init__ = _dft_common_init_
    __post_init__ = _dft_common_post_init_

@util.pytree_node(hf.Traced_Attributes, num_args=1)
class RKS(KohnShamDFT, hf.RHF):
    def __init__(self, mol, xc='LDA,VWN', **kwargs):
        hf.RHF.__init__(self, mol)
        KohnShamDFT.__init__(self, xc)
        self.__dict__.update(kwargs)
        # NOTE this has to be after __dict__ update,
        # otherwise stop_grad(mol) won't work.
        # Currently, no grid response is considered.
        KohnShamDFT.__post_init__(self)

    get_veff = get_veff
    energy_elec = energy_elec
    nuc_grad_method = pyscf_rks.RKS.nuc_grad_method #analytic nuclear gradient method



import numpy
from jax import vmap
from pyscf import lib as pyscf_lib
from pyscf.lib import logger
from pyscf.pbc.tools import get_coulG
from pyscf.pbc.df.df_jk import _format_dms, _format_kpts_band, _format_jks
from pyscf.pbc.lib.kpts_helper import is_zero, gamma_point
from pyscfad.lib import numpy as np
from pyscfad.lib import ops
from pyscfad.pbc import tools
from pyscfad.pbc.df.df_jk import _ewald_exxdiv_for_G0

def get_j_kpts(mydf, dm_kpts, hermi=1, kpts=numpy.zeros((1,3)), kpts_band=None, cell=None):
    if cell is None:
        cell = mydf.cell
    mesh = mydf.mesh

    ni = mydf._numint
    make_rho, nset, nao = ni._gen_rho_evaluator(cell, dm_kpts, hermi)
    dm_kpts = np.asarray(dm_kpts)
    dms = _format_dms(dm_kpts, kpts)
    nset, nkpts, nao = dms.shape[:3]

    coulG = get_coulG(cell, mesh=mesh)
    ngrids = len(coulG)

    kpts_band, input_band = _format_kpts_band(kpts_band, kpts), kpts_band
    nband = len(kpts_band)

    coords = mydf.grids.coords
    mask = mydf.grids.non0tab
    ao2_kpts = mydf._numint.eval_ao(cell, coords, kpts=kpts, non0tab=mask)
    if input_band is None:
        ao1_kpts = ao2_kpts
    else:
        ao1_kpts = mydf._numint.eval_ao(cell, coords, kpts=kpts_band, non0tab=mask)

    if hermi == 1 or gamma_point(kpts):
        #rhoR = np.zeros((nset,ngrids))
        #for ao_ks_etc, p0, p1 in mydf.aoR_loop(mydf.grids, kpts, cell=cell):
        #    ao_ks, mask = ao_ks_etc[0], ao_ks_etc[2]
        #    for i in range(nset):
        #        #rhoR[i,p0:p1] += make_rho(i, ao_ks, mask, 'LDA')
        #        rhoR = ops.index_add(rhoR, ops.index[i,p0:p1],
        #                             make_rho(i, ao_ks, mask, 'LDA'))
        #    ao = ao_ks = None

        rhoR = [make_rho(i, ao2_kpts, mask, 'LDA') for i in range(nset)]
        rhoR = np.asarray(rhoR, dtype=np.double)

        def _rhoR_to_vR_real(rhoR):
            rhoG = tools.fft(rhoR, mesh)
            vG = coulG * rhoG
            vR = tools.ifft(vG, mesh).real
            return vR
        vR = vmap(_rhoR_to_vR_real)(rhoR)

    else:  # vR may be complex if the underlying density is complex
        rhoR = np.zeros((nset,ngrids), dtype=np.complex128)
        #for ao_ks_etc, p0, p1 in mydf.aoR_loop(mydf.grids, kpts, cell=cell):
        #    ao_ks, mask = ao_ks_etc[0], ao_ks_etc[2]
        #    for i in range(nset):
        #        for k, ao in enumerate(ao_ks):
        #            ao_dm = np.dot(ao, dms[i,k])
        #            #rhoR[i,p0:p1] += np.einsum('xi,xi->x', ao_dm, ao.conj())
        #            rhoR = ops.index_add(rhoR, ops.index[i,p0:p1], 
        #                                 np.einsum('xi,xi->x', ao_dm, ao.conj()))
        for i in range(nset):
            for k, ao in enumerate(ao2_kpts):
                ao_dm = np.dot(ao, dms[i,k])
                rhoR = rhoR.at[i].add(np.einsum('xi,xi->x', ao_dm, ao.conj()))
        rhoR *= 1./nkpts

        def _rhoR_to_vR_complex(rhoR):
            rhoG = tools.fft(rhoR, mesh)
            vG = coulG * rhoG
            vR = tools.ifft(vG, mesh)
            return vR
        vR = vmap(_rhoR_to_vR_complex)(rhoR)

    #kpts_band, input_band = _format_kpts_band(kpts_band, kpts), kpts_band
    #nband = len(kpts_band)
    weight = cell.vol / ngrids
    vR *= weight
    if gamma_point(kpts_band):
        vj_kpts = np.zeros((nset,nband,nao,nao))
    else:
        vj_kpts = np.zeros((nset,nband,nao,nao), dtype=np.complex128)

    #for ao_ks_etc, p0, p1 in mydf.aoR_loop(mydf.grids, kpts_band, cell=cell):
    #    ao_ks, mask = ao_ks_etc[0], ao_ks_etc[2]
    #    for i in range(nset):
    #        # ni.eval_mat can handle real vR only
    #        # vj_kpts[i] += ni.eval_mat(cell, ao_ks, 1., None, vR[i,p0:p1], mask, 'LDA')
    #        for k, ao in enumerate(ao_ks):
    #            aow = np.einsum('xi,x->xi', ao, vR[i,p0:p1])
    #            #vj_kpts[i,k] += lib.dot(ao.conj().T, aow)
    #            vj_kpts = ops.index_add(vj_kpts, ops.index[i,k],
    #                                    np.dot(ao.conj().T, aow))

    for i in range(nset):
        for k, ao in enumerate(ao1_kpts):
            aow = np.einsum('xi,x->xi', ao, vR[i])
            vj_kpts = vj_kpts.at[i,k].add(np.dot(ao.conj().T, aow))

    return _format_jks(vj_kpts, dm_kpts, input_band, kpts)

def get_k_kpts(mydf, dm_kpts, hermi=1, kpts=numpy.zeros((1,3)), kpts_band=None,
               exxdiv=None, cell=None):
    if cell is None:
        cell = mydf.cell
    mesh = mydf.mesh
    coords = cell.gen_uniform_grids(mesh)
    ngrids = coords.shape[0]

    if getattr(dm_kpts, 'mo_coeff', None) is not None:
        mo_coeff = dm_kpts.mo_coeff
        mo_occ   = dm_kpts.mo_occ
    else:
        mo_coeff = None

    kpts = np.asarray(kpts)
    dm_kpts = np.asarray(dm_kpts)
    dms = _format_dms(dm_kpts, kpts)
    nset, nkpts, nao = dms.shape[:3]

    weight = 1./nkpts * (cell.vol/ngrids)

    kpts_band, input_band = _format_kpts_band(kpts_band, kpts), kpts_band
    nband = len(kpts_band)

    if gamma_point(kpts_band) and gamma_point(kpts):
        vk_kpts = np.zeros((nset,nband,nao,nao), dtype=dms.dtype)
    else:
        vk_kpts = np.zeros((nset,nband,nao,nao), dtype=np.complex128)

    coords = mydf.grids.coords
    ao2_kpts = [np.asarray(ao.T)
                for ao in mydf._numint.eval_ao(cell, coords, kpts=kpts)]
    if input_band is None:
        ao1_kpts = ao2_kpts
    else:
        ao1_kpts = [np.asarray(ao.T)
                    for ao in mydf._numint.eval_ao(cell, coords, kpts=kpts_band)]
    if mo_coeff is not None and nset == 1:
        mo_coeff = [mo_coeff[k][:,occ>0] * numpy.sqrt(occ[occ>0])
                    for k, occ in enumerate(mo_occ)]
        ao2_kpts = [np.dot(mo_coeff[k].T, ao) for k, ao in enumerate(ao2_kpts)]

    mem_now = pyscf_lib.current_memory()[0]
    max_memory = mydf.max_memory - mem_now
    blksize = int(min(nao, max(1, (max_memory-mem_now)*1e6/16/4/ngrids/nao)))
    logger.debug1(mydf, 'fft_jk: get_k_kpts max_memory %s  blksize %d',
                  max_memory, blksize)
    #ao1_dtype = np.result_type(*ao1_kpts)
    #ao2_dtype = np.result_type(*ao2_kpts)
    vR_dm = np.empty((nset,nao,ngrids), dtype=vk_kpts.dtype)

    #t1 = (logger.process_clock(), logger.perf_counter())
    for k2, ao2T in enumerate(ao2_kpts):
        if ao2T.size == 0:
            continue

        kpt2 = kpts[k2]
        naoj = ao2T.shape[0]
        if mo_coeff is None or nset > 1:
            ao_dms = [np.dot(dms[i,k2], ao2T.conj()) for i in range(nset)]
        else:
            ao_dms = [ao2T.conj()]

        for k1, ao1T in enumerate(ao1_kpts):
            kpt1 = kpts_band[k1]

            # If we have an ewald exxdiv, we add the G=0 correction near the
            # end of the function to bypass any discretization errors
            # that arise from the FFT.
            if exxdiv == 'ewald' or exxdiv is None:
                coulG = get_coulG(cell, kpt2-kpt1, False, mydf, mesh)
            else:
                coulG = get_coulG(cell, kpt2-kpt1, exxdiv, mydf, mesh)
            if is_zero(kpt1-kpt2):
                expmikr = numpy.array(1.)
            else:
                expmikr = np.exp(-1j * np.dot(coords, kpt2-kpt1))

            for p0, p1 in pyscf_lib.prange(0, nao, blksize):
                rho1 = np.einsum('ig,jg->ijg', ao1T[p0:p1].conj()*expmikr, ao2T)
                vG = tools.fft(rho1.reshape(-1,ngrids), mesh)
                rho1 = None
                vG *= coulG
                vR = tools.ifft(vG, mesh).reshape(p1-p0,naoj,ngrids)
                vG = None
                if vR_dm.dtype == np.double:
                    vR = vR.real
                for i in range(nset):
                    #np.einsum('ijg,jg->ig', vR, ao_dms[i], out=vR_dm[i,p0:p1])
                    vR_dm = ops.index_update(vR_dm, ops.index[i,p0:p1],
                                             np.einsum('ijg,jg->ig', vR, ao_dms[i]))
                vR = None
            vR_dm *= expmikr.conj()

            for i in range(nset):
                #vk_kpts[i,k1] += weight * lib.dot(vR_dm[i], ao1T.T)
                vk_kpts = ops.index_add(vk_kpts, ops.index[i,k1],
                                        weight * np.dot(vR_dm[i], ao1T.T))
        #t1 = logger.timer_debug1(mydf, 'get_k_kpts: make_kpt (%d,*)'%k2, *t1)

    # Function _ewald_exxdiv_for_G0 to add back in the G=0 component to vk_kpts
    # Note in the _ewald_exxdiv_for_G0 implementation, the G=0 treatments are
    # different for 1D/2D and 3D systems.  The special treatments for 1D and 2D
    # can only be used with AFTDF/GDF/MDF method.  In the FFTDF method, 1D, 2D
    # and 3D should use the ewald probe charge correction.
    if exxdiv == 'ewald':
        vk_kpts = _ewald_exxdiv_for_G0(cell, kpts, dms, vk_kpts, kpts_band=kpts_band)

    return _format_jks(vk_kpts, dm_kpts, input_band, kpts)

def get_jk(mydf, dm, hermi=1, kpt=numpy.zeros(3), kpts_band=None,
           with_j=True, with_k=True, exxdiv=None, cell=None):
    dm = np.asarray(dm)
    vj = vk = None
    if with_j:
        vj = get_j(mydf, dm, hermi, kpt, kpts_band, cell=cell)
    if with_k:
        vk = get_k(mydf, dm, hermi, kpt, kpts_band, exxdiv, cell=cell)
    return vj, vk

def get_j(mydf, dm, hermi=1, kpt=numpy.zeros(3), kpts_band=None, cell=None):
    dm = np.asarray(dm)
    nao = dm.shape[-1]
    dm_kpts = dm.reshape(-1,1,nao,nao)
    vj = get_j_kpts(mydf, dm_kpts, hermi, kpt.reshape(1,3), kpts_band, cell=cell)
    if kpts_band is None:
        vj = vj[:,0]
    if dm.ndim == 2:
        vj = vj[0]
    return vj

def get_k(mydf, dm, hermi=1, kpt=numpy.zeros(3), kpts_band=None, exxdiv=None, cell=None):
    dm = np.asarray(dm)
    nao = dm.shape[-1]
    dm_kpts = dm.reshape(-1,1,nao,nao)
    vk = get_k_kpts(mydf, dm_kpts, hermi, kpt.reshape(1,3), kpts_band, exxdiv, cell=cell)
    if kpts_band is None:
        vk = vk[:,0]
    if dm.ndim == 2:
        vk = vk[0]
    return vk



from pyscf.lib import logger, split_reshape
from pyscf.mp import mp2 as pyscf_mp2
from pyscfad import util
from pyscfad import lib
from pyscfad.lib import numpy as np

# Iteratively solve MP2 if non-canonical HF is provided
def _iterative_kernel(mp, eris, verbose=None):
    cput1 = cput0 = (logger.process_clock(), logger.perf_counter())
    log = logger.new_logger(mp, verbose)

    emp2, t2 = mp.init_amps(eris=eris)
    log.info('Init E(MP2) = %.15g', emp2)

    adiis = lib.diis.DIIS(mp)

    conv = False
    for istep in range(mp.max_cycle):
        t2new = mp.update_amps(t2, eris)

        if hasattr(t2new, 'ndim'):
            normt = np.linalg.norm(t2new - t2)
            t2 = None
            t2new = adiis.update(t2new)
        else: # UMP2
            normt = np.linalg.norm([np.linalg.norm(t2new[i] - t2[i])
                                     for i in range(3)])
            t2 = None
            t2shape = [x.shape for x in t2new]
            t2new = np.hstack([x.ravel() for x in t2new])
            t2new = adiis.update(t2new)
            t2new = split_reshape(t2new, t2shape)

        t2, t2new = t2new, None
        emp2, e_last = mp.energy(t2, eris), emp2
        log.info('cycle = %d  E_corr(MP2) = %.15g  dE = %.9g  norm(t2) = %.6g',
                 istep+1, emp2, emp2 - e_last, normt)
        cput1 = log.timer('MP2 iter', *cput1)
        if abs(emp2-e_last) < mp.conv_tol and normt < mp.conv_tol_normt:
            conv = True
            break
    log.timer('MP2', *cput0)
    del log
    return conv, emp2, t2

@util.pytree_node(['_scf', 'mol'], num_args=1)
class MP2(pyscf_mp2.MP2):
    def __init__(self, mf, frozen=None, mo_coeff=None, mo_occ=None, **kwargs):
        pyscf_mp2.MP2.__init__(self, mf, frozen=frozen,
                               mo_coeff=mo_coeff, mo_occ=mo_occ)
        self.__dict__.update(kwargs)

    def ao2mo(self, mo_coeff=None):
        eris = pyscf_mp2._ChemistsERIs()
        eris._common_init_(self, mo_coeff)
        mo_coeff = eris.mo_coeff

        nocc = self.nocc
        co = np.asarray(mo_coeff[:,:nocc])
        cv = np.asarray(mo_coeff[:,nocc:])
        eris.ovov = np.einsum("uvst,ui,va,sj,tb->iajb", self._scf._eri, co,cv,co,cv)
        return eris

    _iterative_kernel = _iterative_kernel



import os
from setuptools import setup, find_packages, Extension
from setuptools.command.build_py import build_py
from setuptools.command.build_ext import build_ext

_dct = {}
with open('pyscfad/version.py') as f:
  exec(f.read(), _dct)
__version__ = _dct['__version__']

class CMakeBuildExt(build_ext):
    def run(self):
        extension = self.extensions[0]
        assert extension.name == 'pyscfad_lib_placeholder'
        self.build_cmake(extension)

    def build_cmake(self, extension):
        self.announce('Configuring extensions', level=3)
        src_dir = os.path.abspath(os.path.join(__file__, '..', 'pyscfad', 'lib'))
        cmd = ['cmake', f'-S{src_dir}', f'-B{self.build_temp}']
        configure_args = os.getenv('CMAKE_CONFIGURE_ARGS')
        if configure_args:
            cmd.extend(configure_args.split(' '))
        self.spawn(cmd)

        self.announce('Building binaries', level=3)
        cmd = ['cmake', '--build', self.build_temp, '-j']
        build_args = os.getenv('CMAKE_BUILD_ARGS')
        if build_args:
            cmd.extend(build_args.split(' '))
        if self.dry_run:
            self.announce(' '.join(cmd))
        else:
            self.spawn(cmd)

    # To remove the infix string like cpython-37m-x86_64-linux-gnu.so
    # Python ABI updates since 3.5
    # https://www.python.org/dev/peps/pep-3149/
    def get_ext_filename(self, ext_name):
        ext_path = ext_name.split('.')
        filename = build_ext.get_ext_filename(self, ext_name)
        name, ext_suffix = os.path.splitext(filename)
        return os.path.join(*ext_path) + ext_suffix

#from distutils.command.build import build
#build.sub_commands = ([c for c in build.sub_commands if c[0] == 'build_ext'] +
#                      [c for c in build.sub_commands if c[0] != 'build_ext'])

class BuildExtFirst(build_py):
    def run(self):
        self.run_command("build_ext")
        return super().run()

setup(
    name='pyscfad',
    version=__version__,
    description='PySCF with autodiff',
    author='Xing Zhang',
    author_email='xzhang8@caltech.edu',
    include_package_data=True,
    packages=find_packages(exclude=["examples","*test*"]),
    python_requires='>=3.7',
    #ext_modules=[Extension('pyscfad_lib_placeholder', [])],
    #cmdclass={'build_py': BuildExtFirst,
    #          'build_ext': CMakeBuildExt},
    install_requires=[
        'numpy>=1.17',
        'scipy',
        'jax>=0.1.65,<0.3.14',
        'jaxlib>=0.1.65,<0.3.14',
        'jaxopt>=0.2',
#        'pyscf @ git+https://github.com/fishjojo/pyscf.git@ad#egg=pyscf',
#        'pyscf-properties @ git+https://github.com/fishjojo/properties.git@ad',
    ],
    url='https://github.com/fishjojo/pyscfad',
    license='MIT',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Scientific/Engineering",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    zip_safe=False,
)



from jax import vmap
from pyscf import __config__
from pyscf import numpy as np
from pyscf import gto
from pyscf.lib import logger, param
from pyscf.gto.mole import PTR_ENV_START
from pyscfad import util
from pyscfad.lib import ops, custom_jvp
from pyscfad.gto import moleintor
from pyscfad.gto.eval_gto import eval_gto
from ._mole_helper import setup_exp, setup_ctr_coeff

Traced_Attributes = ['coords', 'exp', 'ctr_coeff', 'r0']

def energy_nuc(mol, charges=None, **kwargs):
    if charges is None:
        charges = mol.atom_charges()
    if len(charges) <= 1:
        return 0.0
    rr = distance_matrix(mol.atom_coords())
    enuc = np.einsum('i,ij,j->', charges, 1./rr, charges) * .5
    return enuc

@custom_jvp
def distance_matrix(coords):
    rr  = np.linalg.norm(coords[:,None,:] - coords[None,:,:], axis=2)
    rr += np.eye(rr.shape[-1]) * 1e200
    return rr

@distance_matrix.defjvp
def distance_matrix_jvp(primals, tangents):
    coords, = primals
    coords_t, = tangents
    rnorm = primal_out = distance_matrix(coords)
    def body(r1, r2, rnorm, coords_t):
        r = r1 - r2
        jvp = np.dot(r / rnorm[:,None], coords_t)
        return jvp
    tangent_out = vmap(body, (0,None,0,0))(coords, coords, rnorm, coords_t)
    tangent_out += tangent_out.T
    return primal_out, tangent_out

intor_cross = moleintor.intor_cross

@util.pytree_node(Traced_Attributes)
class Mole(gto.Mole):
    '''
    A subclass of :class:`pyscf.gto.Mole`, where the following
    attributes can be traced.

    Attributes:
        coords : array
            Atomic coordinates.
        exp : array
            Exponents of Gaussian basis functions.
        ctr_coeff : array
            Contraction coefficients of Gaussian basis functions.
        r0 : array
            Centers of Gaussian basis functions. Currently this is
            not used as the basis functions are atom centered. This
            is a placeholder for floating Gaussian basis sets.
    '''
    def __init__(self, **kwargs):
        self.coords = None
        self.exp = None
        self.ctr_coeff = None
        self.r0 = None
        gto.Mole.__init__(self, **kwargs)

    def atom_coords(self, unit='Bohr'):
        if self.coords is None:
            return gto.Mole.atom_coords(self, unit)
        else:
            if unit[:3].upper() == 'ANG':
                return self.coords * param.BOHR
            else:
                return self.coords

    def build(self, *args, **kwargs):
        trace_coords = kwargs.pop("trace_coords", True)
        trace_exp = kwargs.pop("trace_exp", True)
        trace_ctr_coeff = kwargs.pop("trace_ctr_coeff", True)
        trace_r0 = kwargs.pop("trace_r0", False)

        gto.Mole.build(self, *args, **kwargs)

        if trace_coords:
            self.coords = np.asarray(self.atom_coords())
        if trace_exp:
            self.exp, _, _ = setup_exp(self)
        if trace_ctr_coeff:
            self.ctr_coeff, _, _ = setup_ctr_coeff(self)
        if trace_r0:
            raise NotImplementedError

    energy_nuc = energy_nuc
    eval_ao = eval_gto = eval_gto

    def intor(self, intor, comp=None, hermi=0, aosym='s1', out=None,
              shls_slice=None, grids=None):
        if (self.coords is None and self.exp is None
                and self.ctr_coeff is None and self.r0 is None):
            return gto.Mole.intor(self, intor, comp=comp, hermi=hermi,
                                  aosym=aosym, out=out, shls_slice=shls_slice)
        else:
            return moleintor.getints(self, intor, shls_slice,
                                     comp, hermi, aosym, out=None)



from . import df
from . import addons
from .df import DF

def density_fit(mf, auxbasis=None, with_df=None, only_dfj=False):
    # pylint: disable=import-outside-toplevel
    from pyscfad import util
    from pyscf.df import df_jk as pyscf_df_jk
    if with_df is None:
        #auxmol has to be defined before creating the DF object,
        #otherwise, its derivative won't be traced
        auxmol = addons.make_auxmol(mf.mol, auxbasis)
        with_df = DF(mf.mol, auxmol=auxmol)
        with_df.max_memory = mf.max_memory
        with_df.stdout = mf.stdout
        with_df.verbose = mf.verbose
        with_df.auxbasis = auxbasis

    mf_class = mf.__class__
    kwargs = mf.__dict__.copy()
    mol = kwargs.pop("mol")

    # pylint: disable=abstract-method
    @util.pytree_node(['mol', 'with_df'], num_args=1)
    class DFHF(pyscf_df_jk._DFHF, mf_class):
        def __init__(self, mol, with_df=None, only_dfj=False, **kwargs):
            self.with_df = with_df
            if getattr(self.with_df, "mol", None) is not None:
                self.mol = self.with_df.mol
            else:
                self.mol = mol
            self.only_dfj = only_dfj
            self.__dict__.update(kwargs)

            self._eri = None
            self._keys = self._keys.union(['with_df', 'only_dfj'])

        def reset(self, mol=None):
            self.with_df.reset(mol)
            return mf_class.reset(self, mol)

        def get_jk(self, mol=None, dm=None, hermi=1, with_j=True, with_k=True,
                   omega=None):
            if dm is None:
                dm = self.make_rdm1()
            if self.with_df and self.only_dfj:
                vj = vk = None
                if with_j:
                    vj, vk = self.with_df.get_jk(dm, hermi, True, False,
                                                 self.direct_scf_tol, omega)
                if with_k:
                    vk = mf_class.get_jk(self, mol, dm, hermi, False, True, omega)[1]
            elif self.with_df:
                vj, vk = self.with_df.get_jk(dm, hermi, with_j, with_k,
                                             self.direct_scf_tol, omega)
            else:
                vj, vk = mf_class.get_jk(self, mol, dm, hermi, with_j, with_k, omega)
            return vj, vk

    return DFHF(mol, with_df, only_dfj, **kwargs)



import numpy
from jax import vmap
from pyscf.lib import direct_sum
from pyscf.tdscf import rhf as pyscf_tdrhf
from pyscfad import util
from pyscfad import ao2mo
from pyscfad.lib import numpy as np
from pyscfad.gto import mole

Traced_Attributes = ['_scf', 'mol']

def get_ab(mf, mo_energy=None, mo_coeff=None, mo_occ=None):
    r'''A and B matrices for TDDFT response function.

    A[i,a,j,b] = \delta_{ab}\delta_{ij}(E_a - E_i) + (ia||bj)
    B[i,a,j,b] = (ia||jb)
    '''
    if mo_energy is None:
        mo_energy = mf.mo_energy
    if mo_coeff is None:
        mo_coeff = mf.mo_coeff
    if mo_occ is None:
        mo_occ = mf.mo_occ
    assert mo_coeff.dtype == np.double

    nmo = mo_coeff.shape[-1]
    occidx = np.where(mo_occ==2)[0]
    viridx = np.where(mo_occ==0)[0]
    orbv = mo_coeff[:,viridx]
    orbo = mo_coeff[:,occidx]
    nvir = orbv.shape[1]
    nocc = orbo.shape[1]
    mo = np.hstack((orbo,orbv))
    nmo = nocc + nvir

    e_ia = direct_sum('a-i->ia', mo_energy[viridx], mo_energy[occidx])
    a = np.diag(e_ia.ravel()).reshape(nocc,nvir,nocc,nvir)
    b = np.zeros_like(a)

    def add_hf(a, b, hyb=1):
        eri_mo = ao2mo.incore.general(mf._eri, [orbo,mo,mo,mo], compact=False)
        eri_mo = eri_mo.reshape(nocc,nmo,nmo,nmo)
        a  = np.einsum('iabj->iajb', eri_mo[:nocc,nocc:,nocc:,:nocc]) * 2
        a -= np.einsum('ijba->iajb', eri_mo[:nocc,:nocc,nocc:,nocc:]) * hyb

        b  = np.einsum('iajb->iajb', eri_mo[:nocc,nocc:,:nocc,nocc:]) * 2
        b -= np.einsum('jaib->iajb', eri_mo[:nocc,nocc:,:nocc,nocc:]) * hyb
        return a, b

    a_hf, b_hf = add_hf(a,b)
    a += a_hf
    b += b_hf
    return a, b

def cis_ovlp(mol1, mol2, mo1, mo2, nocc1, nocc2, nmo1, nmo2, x1, x2):
    s_ao = mole.intor_cross('int1e_ovlp', mol1, mol2)
    nvir1 = nmo1 - nocc1

    idx1 = []
    for i in range(nocc1):
        for a in range(nocc1,nmo1):
            idx = numpy.arange(nocc1)
            idx[i] = a
            idx1.append(idx)
    idx1 = np.asarray(idx1)

    if nocc1 == nocc2 and nmo1 == nmo2:
        idx2 = idx1
    else:
        idx2 = []
        for j in range(nocc2):
            for b in range(nocc2,nmo2):
                idx = numpy.arange(nocc2)
                idx[j] = b
                idx2.append(idx)
        idx2 = np.asarray(idx2)

    def body(idx, mo1_occ):
        s_mo = np.einsum('ui,uv,vj->ij', mo1_occ, s_ao, mo2[:,idx])
        return np.linalg.det(s_mo)

    res = 0.
    for i in range(nocc1):
        for a in range(nvir1):
            mo1_occ = mo1[:,idx1[i*nvir1+a]]
            s12 = vmap(body, (0,None))(idx2, mo1_occ)
            res += x1[i,a] * (s12 * x2.ravel()).sum()
    return res

# pylint: disable=abstract-method
@util.pytree_node(Traced_Attributes, num_args=1)
class TDMixin(pyscf_tdrhf.TDMixin):
    def __init__(self, mf, **kwargs):
        pyscf_tdrhf.TDMixin.__init__(self, mf)
        self.__dict__.update(kwargs)

    def get_ab(self, mf=None):
        if mf is None:
            mf = self._scf
        return get_ab(mf)

@util.pytree_node(Traced_Attributes, num_args=1)
class TDA(TDMixin, pyscf_tdrhf.TDA):
    pass

CIS = TDA



import warnings
from functools import partial
import numpy
from jax import jit, vmap, custom_jvp
from pyscf import numpy as np
from pyscf.lib import load_library
from pyscf.dft import numint
from pyscf.dft.numint import SWITCH_SIZE
from pyscf.dft.gen_grid import make_mask, BLKSIZE
from pyscfad.lib import ops
from pyscfad.lib import stop_grad
from pyscfad.dft import libxc

libdft = load_library('libdft')

def eval_mat(mol, ao, weight, rho, vxc,
             non0tab=None, xctype='LDA', spin=0, verbose=None):
    xctype = xctype.upper()
    if xctype in ['LDA', 'HF']:
        ngrids, _ = ao.shape
    else:
        ngrids, _ = ao[0].shape

    if non0tab is None:
        non0tab = numpy.ones(((ngrids+BLKSIZE-1)//BLKSIZE,mol.nbas),
                             dtype=numpy.uint8)
    shls_slice = (0, mol.nbas)
    ao_loc = mol.ao_loc_nr()
    transpose_for_uks = False
    if xctype in ['LDA', 'HF']:
        if not getattr(vxc, 'ndim', None) == 2:
            vrho = vxc[0]
        else:
            vrho = vxc
        # *.5 because return mat + mat.T
        #:aow = numpy.einsum('pi,p->pi', ao, .5*weight*vrho)
        aow = _scale_ao(ao, .5*weight*vrho)
        mat = _dot_ao_ao(mol, ao, aow, non0tab, shls_slice, ao_loc)
    else:
        vrho, vsigma = vxc[:2]
        if spin == 0:
            assert(vsigma is not None and rho.ndim==2)
            wv_rho = weight * vrho * .5
            wv_sigma = rho[1:4] * (weight * vsigma * 2)
            wv = np.concatenate((wv_rho.reshape(1,-1), wv_sigma))
        else:
            rho_a, rho_b = rho
            wv_rho = weight * vrho * .5
            try:
                wv_sigma  = rho_a[1:4] * (weight * vsigma[0] * 2)
                wv_sigma += rho_b[1:4] * (weight * vsigma[1])
            except ValueError:
                warnings.warn('Note the output of libxc.eval_xc cannot be '
                              'directly used in eval_mat.\nvsigma from eval_xc '
                              'should be restructured as '
                              '(vsigma[:,0],vsigma[:,1])\n')
                transpose_for_uks = True
                vsigma = vsigma.T
                wv_sigma  = rho_a[1:4] * (weight * vsigma[0] * 2)
                wv_sigma += rho_b[1:4] * (weight * vsigma[1])
            wv = np.concatenate((wv_rho.reshape(1,-1), wv_sigma))
        #:aow = numpy.einsum('npi,np->pi', ao[:4], wv)
        aow = _scale_ao(ao[:4], wv)
        mat = _dot_ao_ao(mol, ao[0], aow, non0tab, shls_slice, ao_loc)

# JCP 138, 244108 (2013); DOI:10.1063/1.4811270
# JCP 112, 7002 (2000); DOI:10.1063/1.481298
    if xctype == 'MGGA':
        vlapl, vtau = vxc[2:]

        if vlapl is None:
            vlapl = 0
        else:
            if spin != 0:
                if transpose_for_uks:
                    vlapl = vlapl.T
                vlapl = vlapl[0]
            XX, YY, ZZ = 4, 7, 9
            ao2 = ao[XX] + ao[YY] + ao[ZZ]
            #:aow = numpy.einsum('pi,p->pi', ao2, .5 * weight * vlapl, out=aow)
            aow = _scale_ao(ao2, .5 * weight * vlapl, out=None)
            mat += _dot_ao_ao(mol, ao[0], aow, non0tab, shls_slice, ao_loc)

        if spin != 0:
            if transpose_for_uks:
                vtau = vtau.T
            vtau = vtau[0]
        wv = weight * (.25*vtau + vlapl)
        #:aow = numpy.einsum('pi,p->pi', ao[1], wv, out=aow)
        aow = _scale_ao(ao[1], wv, out=None)
        mat += _dot_ao_ao(mol, ao[1], aow, non0tab, shls_slice, ao_loc)
        #:aow = numpy.einsum('pi,p->pi', ao[2], wv, out=aow)
        aow = _scale_ao(ao[2], wv, out=None)
        mat += _dot_ao_ao(mol, ao[2], aow, non0tab, shls_slice, ao_loc)
        #:aow = numpy.einsum('pi,p->pi', ao[3], wv, out=aow)
        aow = _scale_ao(ao[3], wv, out=None)
        mat += _dot_ao_ao(mol, ao[3], aow, non0tab, shls_slice, ao_loc)

    return mat + mat.T.conj()

def nr_rks(ni, mol, grids, xc_code, dms, relativity=0, hermi=0,
           max_memory=2000, verbose=None):
    xctype = ni._xc_type(xc_code)
    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dms, hermi)

    shls_slice = (0, mol.nbas)
    ao_loc = mol.ao_loc_nr()

    nelec = [0]*nset
    excsum = [0]*nset
    vmat = [0]*nset
    aow = None
    if xctype == 'LDA':
        ao_deriv = 0
        for ao, mask, weight, coords \
                in ni.block_loop(mol, grids, nao, ao_deriv, max_memory):
            #aow = numpy.ndarray(ao.shape, order='F', buffer=aow)
            for idm in range(nset):
                rho = make_rho(idm, ao, mask, 'LDA')
                exc, vxc = ni.eval_xc(xc_code, rho, spin=0,
                                      relativity=relativity, deriv=1,
                                      verbose=verbose)[:2]
                vrho = vxc[0]
                den = rho * weight
                nelec[idm] += stop_grad(den).sum()
                excsum[idm] += np.dot(den, exc)
                # *.5 because vmat + vmat.T
                #:aow = numpy.einsum('pi,p->pi', ao, .5*weight*vrho, out=aow)
                aow = _scale_ao(ao, .5*weight*vrho, out=None)
                vmat[idm] += _dot_ao_ao(mol, ao, aow, mask, shls_slice, ao_loc)
                rho = exc = vxc = vrho = None
    elif xctype == 'GGA':
        ao_deriv = 1
        for ao, mask, weight, coords \
                in ni.block_loop(mol, grids, nao, ao_deriv, max_memory):
            #aow = numpy.ndarray(ao[0].shape, order='F', buffer=aow)
            for idm in range(nset):
                rho = make_rho(idm, ao, mask, 'GGA')
                exc, vxc = ni.eval_xc(xc_code, rho, spin=0,
                                      relativity=relativity, deriv=1,
                                      verbose=verbose)[:2]
                den = rho[0] * weight
                nelec[idm] += stop_grad(den).sum()
                excsum[idm] += np.dot(den, exc)
                # ref eval_mat function
                wv = _rks_gga_wv0(rho, vxc, weight)
                #:aow = numpy.einsum('npi,np->pi', ao, wv, out=aow)
                aow = _scale_ao(ao, wv, out=None)
                vmat[idm] += _dot_ao_ao(mol, ao[0], aow, mask, shls_slice, ao_loc)
                rho = exc = vxc = wv = None
    elif xctype == 'NLC':
        nlc_pars = ni.nlc_coeff(xc_code[:-6])
        if nlc_pars == [0,0]:
            raise NotImplementedError('VV10 cannot be used with %s. '
                                      'The supported functionals are %s' %
                                      (xc_code[:-6], ni.libxc.VV10_XC))
        ao_deriv = 1
        vvrho=numpy.empty([nset,4,0])
        vvweight=numpy.empty([nset,0])
        vvcoords=numpy.empty([nset,0,3])
        for ao, mask, weight, coords \
                in ni.block_loop(mol, grids, nao, ao_deriv, max_memory):
            ao = stop_grad(ao)
            rhotmp = numpy.empty([0,4,weight.size])
            weighttmp = numpy.empty([0,weight.size])
            coordstmp = numpy.empty([0,weight.size,3])
            for idm in range(nset):
                rho = make_rho(idm, ao, mask, 'GGA')
                rho = numpy.asarray(stop_grad(rho))
                rho = numpy.expand_dims(rho,axis=0)
                rhotmp = numpy.concatenate((rhotmp,rho),axis=0)
                weighttmp = numpy.concatenate((weighttmp,numpy.expand_dims(weight,axis=0)),axis=0)
                coordstmp = numpy.concatenate((coordstmp,numpy.expand_dims(coords,axis=0)),axis=0)
                rho = None
            vvrho = numpy.concatenate((vvrho,rhotmp),axis=2)
            vvweight = numpy.concatenate((vvweight,weighttmp),axis=1)
            vvcoords = numpy.concatenate((vvcoords,coordstmp),axis=1)
            rhotmp = weighttmp = coordstmp = None
        for ao, mask, weight, coords \
                in ni.block_loop(mol, grids, nao, ao_deriv, max_memory):
            #aow = numpy.ndarray(ao[0].shape, order='F', buffer=aow)
            for idm in range(nset):
                rho = make_rho(idm, ao, mask, 'GGA')
                exc, vxc = _vv10nlc(rho,coords,vvrho[idm],vvweight[idm],vvcoords[idm],nlc_pars)
                den = rho[0] * weight
                nelec[idm] += stop_grad(den).sum()
                excsum[idm] += np.dot(den, exc)
                # ref eval_mat function
                wv = _rks_gga_wv0(rho, vxc, weight)
                #:aow = numpy.einsum('npi,np->pi', ao, wv, out=aow)
                aow = _scale_ao(ao, wv, out=None)
                vmat[idm] += _dot_ao_ao(mol, ao[0], aow, mask, shls_slice, ao_loc)
                rho = exc = vxc = wv = None
        vvrho = vvweight = vvcoords = None
    elif xctype == 'MGGA':
        if any(x in xc_code.upper() for x in ('CC06', 'CS', 'BR89', 'MK00')):
            raise NotImplementedError('laplacian in meta-GGA method')
        ao_deriv = 2
        for ao, mask, weight, coords \
                in ni.block_loop(mol, grids, nao, ao_deriv, max_memory):
            #aow = numpy.ndarray(ao[0].shape, order='F', buffer=aow)
            for idm in range(nset):
                rho = make_rho(idm, ao, mask, 'MGGA')
                exc, vxc = ni.eval_xc(xc_code, rho, spin=0,
                                      relativity=relativity, deriv=1,
                                      verbose=verbose)[:2]
                # pylint: disable=W0612
                vrho, vsigma, vlapl, vtau = vxc[:4]
                den = rho[0] * weight
                nelec[idm] += stop_grad(den).sum()
                excsum[idm] += np.dot(den, exc)

                wv = _rks_gga_wv0(rho, vxc, weight)
                #:aow = numpy.einsum('npi,np->pi', ao[:4], wv, out=aow)
                aow = _scale_ao(ao[:4], wv, out=None)
                vmat[idm] += _dot_ao_ao(mol, ao[0], aow, mask, shls_slice, ao_loc)
# pylint: disable=W0511
# FIXME: .5 * .5   First 0.5 for v+v.T symmetrization.
# Second 0.5 is due to the Libxc convention tau = 1/2 \nabla\phi\dot\nabla\phi
                wv = (.5 * .5 * weight * vtau).reshape(-1,1)
                vmat[idm] += _dot_ao_ao(mol, ao[1], wv*ao[1], mask, shls_slice, ao_loc)
                vmat[idm] += _dot_ao_ao(mol, ao[2], wv*ao[2], mask, shls_slice, ao_loc)
                vmat[idm] += _dot_ao_ao(mol, ao[3], wv*ao[3], mask, shls_slice, ao_loc)
                rho = exc = vxc = vrho = wv = None

    for i in range(nset):
        vmat[i] = vmat[i] + vmat[i].conj().T
    nelec = numpy.asarray(nelec)
    excsum = np.asarray(excsum)
    vmat = np.asarray(vmat)
    if nset == 1:
        nelec = nelec[0]
        excsum = excsum[0]
        vmat = vmat[0]
    return nelec, excsum, vmat

def eval_rho(mol, ao, dm, non0tab=None, xctype='LDA', hermi=0, verbose=None):
    xctype = xctype.upper()
    if xctype in ('LDA', 'HF'):
        ngrids = ao.shape[0]
    else:
        ngrids = ao[0].shape[0]

    if non0tab is None:
        non0tab = numpy.ones(((ngrids+BLKSIZE-1)//BLKSIZE,mol.nbas),
                             dtype=numpy.uint8)
    if not hermi:
        # (D + D.T)/2 because eval_rho computes 2*(|\nabla i> D_ij <j|) instead of
        # |\nabla i> D_ij <j| + |i> D_ij <\nabla j| for efficiency
        dm = (dm + dm.conj().T) * .5

    shls_slice = (0, mol.nbas)
    ao_loc = mol.ao_loc_nr()
    if xctype in ('LDA', 'HF'):
        c0 = _dot_ao_dm(mol, ao, dm, non0tab, shls_slice, ao_loc)
        #:rho = numpy.einsum('pi,pi->p', ao, c0)
        rho = _contract_rho(ao, c0)
    elif xctype in ('GGA', 'NLC'):
        rho = _rks_gga_assemble_rho(ao, dm)
    else: # meta-GGA
        # rho[4] = \nabla^2 rho, rho[5] = 1/2 |nabla f|^2
        rho = np.empty((6,ngrids))
        #c0 = _dot_ao_dm(mol, ao[0], dm, non0tab, shls_slice, ao_loc)
        #:rho[0] = numpy.einsum('pi,pi->p', ao[0], c0)
        #rho = ops.index_update(rho, ops.index[0], _contract_rho(ao[0], c0))
        #rho = ops.index_update(rho, ops.index[5], 0)
        #for i in range(1, 4):
        #    #:rho[i] = numpy.einsum('pi,pi->p', c0, ao[i]) * 2 # *2 for +c.c.
        #    rho = ops.index_update(rho, ops.index[i], _contract_rho(c0, ao[i]) * 2)
        #    c1 = _dot_ao_dm(mol, ao[i], dm.T, non0tab, shls_slice, ao_loc)
        #    #:rho[5] += numpy.einsum('pi,pi->p', c1, ao[i])
        #    rho = ops.index_add(rho, ops.index[5], _contract_rho(c1, ao[i]))
        #XX, YY, ZZ = 4, 7, 9
        #ao2 = ao[XX] + ao[YY] + ao[ZZ]
        ##:rho[4] = numpy.einsum('pi,pi->p', c0, ao2)
        #rho = ops.index_update(rho, ops.index[4], _contract_rho(c0, ao2))
        #rho = ops.index_add(rho, ops.index[4], rho[5])
        #rho = ops.index_mul(rho, ops.index[4], 2)
        #rho = ops.index_mul(rho, ops.index[5], .5)
        rho = _rks_mgga_assemble_rho(rho, ao, dm)
    return rho

@jit
def _rks_gga_assemble_rho(ao, dm):
    c0 = _dot_ao_dm_incore(ao[0], dm)
    factor = np.asarray([1., 2., 2., 2.])
    rho = vmap(_contract_rho, (None,0,0))(c0, ao, factor)
    return rho

@jit
def _rks_mgga_assemble_rho(rho, ao, dm):
    c0 = _dot_ao_dm_incore(ao[0], dm)
    rho = ops.index_update(rho, ops.index[0], _contract_rho(ao[0], c0))
    rho = ops.index_update(rho, ops.index[5], 0)
    for i in range(1, 4):
        rho = ops.index_update(rho, ops.index[i], _contract_rho(c0, ao[i]) * 2)
        c1 = _dot_ao_dm_incore(ao[i], dm.T)
        rho = ops.index_add(rho, ops.index[5], _contract_rho(c1, ao[i]))
    XX, YY, ZZ = 4, 7, 9
    ao2 = ao[XX] + ao[YY] + ao[ZZ]
    rho = ops.index_update(rho, ops.index[4], _contract_rho(c0, ao2))
    rho = ops.index_add(rho, ops.index[4], rho[5])
    rho = ops.index_mul(rho, ops.index[4], 2)
    rho = ops.index_mul(rho, ops.index[5], .5)
    return rho

@jit
def _scale_ao(ao, wv, out=None):
    #:aow = numpy.einsum('npi,np->pi', ao[:4], wv)
    if wv.ndim == 2:
        ao = ao.transpose(0,2,1)
    else:
        ngrids, nao = ao.shape
        ao = ao.T.reshape(1,nao,ngrids)
        wv = wv.reshape(1,ngrids)

    aow = np.einsum('nip,np->pi', ao, wv)
    return aow

def _dot_ao_ao(mol, ao1, ao2, non0tab, shls_slice, ao_loc, hermi=0):
    '''return numpy.dot(ao1.T, ao2)'''
    nao = ao1.shape[-1]
    if nao < SWITCH_SIZE:
        return _dot_ao_ao_incore(ao1, ao2)
    else:
        raise NotImplementedError

@jit
def _dot_ao_ao_incore(ao1, ao2):
    return np.dot(ao1.T.conj(), ao2)

def _dot_ao_dm(mol, ao, dm, non0tab, shls_slice, ao_loc, out=None):
    '''return numpy.dot(ao, dm)'''
    nao = ao.shape[-1]
    if nao < SWITCH_SIZE:
        return _dot_ao_dm_incore(ao, dm)
    else:
        raise NotImplementedError

@jit
def _dot_ao_dm_incore(ao, dm):
    return np.dot(np.asarray(dm).T, ao.T).T

@jit
def _contract_rho(bra, ket, factor=1.0):
    bra = bra.T
    ket = ket.T

    rho  = np.einsum('ip,ip->p', bra.real, ket.real)
    rho += np.einsum('ip,ip->p', bra.imag, ket.imag)
    return rho * factor

@jit
def _rks_gga_wv0(rho, vxc, weight):
    vrho, vgamma = vxc[:2]
    wv_rho = weight * vrho * .5
    wv_sigma = (weight * vgamma * 2) * rho[1:4]
    wv = np.concatenate((wv_rho.reshape(1,-1), wv_sigma))
    #wv = ops.index_mul(wv, ops.index[0], .5)  # v+v.T should be applied in the caller
    return wv

@partial(custom_jvp, nondiff_argnums=(1,2,3,4,5,))
def _vv10nlc(rho, coords, vvrho, vvweight, vvcoords, nlc_pars):
    rho = numpy.asarray(rho)
    return numint._vv10nlc(rho, coords, vvrho, vvweight, vvcoords, nlc_pars)

@_vv10nlc.defjvp
def _vv10nlc_jvp(coords, vvrho, vvweight, vvcoords, nlc_pars,
                 primals, tangents):
    rho, = primals
    rho_t, = tangents

    exc, vxc = _vv10nlc(rho, coords, vvrho, vvweight, vvcoords, nlc_pars)

    exc_jvp = (vxc[0] - exc) / rho[0] * rho_t[0] \
            + vxc[1] / rho[0] * 2. * np.einsum('np,np->p', rho[1:4], rho_t[1:4])
    # pylint: disable=W0511
    vxc_jvp = np.zeros_like(vxc) # FIXME gradient of vxc not implemented
    return (exc,vxc), (exc_jvp, vxc_jvp)

class NumInt(numint.NumInt):
    def _gen_rho_evaluator(self, mol, dms, hermi=0):
        if getattr(dms, 'mo_coeff', None) is not None:
            # pylint: disable=W0511
            #TODO: test whether dm.mo_coeff matching dm
            mo_coeff = dms.mo_coeff
            mo_occ = dms.mo_occ
            if isinstance(dms, numpy.ndarray) and dms.ndim == 2:
                mo_coeff = [mo_coeff]
                mo_occ = [mo_occ]
            nao = mo_coeff[0].shape[0]
            ndms = len(mo_occ)
            def make_rho(idm, ao, non0tab, xctype):
                return self.eval_rho2(mol, ao, mo_coeff[idm], mo_occ[idm],
                                      non0tab, xctype)
        else:
            if getattr(dms, "ndim", None) == 2:
                dms = [dms]
            if not hermi:
                # For eval_rho when xctype==GGA, which requires hermitian DMs
                dms = [(dm+dm.conj().T)*.5 for dm in dms]
            nao = dms[0].shape[0]
            ndms = len(dms)
            def make_rho(idm, ao, non0tab, xctype):
                return self.eval_rho(mol, ao, dms[idm], non0tab, xctype, hermi=1)
        return make_rho, ndms, nao

    def eval_xc(self, xc_code, rho, spin=0, relativity=0, deriv=1, omega=None,
                verbose=None):
        if omega is None:
            omega = self.omega
        return libxc.eval_xc(xc_code, rho, spin, relativity, deriv,
                             omega, verbose)

    def eval_rho(self, mol, ao, dm, non0tab=None, xctype='LDA', hermi=0, verbose=None):
        return eval_rho(mol, ao, dm, non0tab, xctype, hermi, verbose)

    nr_rks = nr_rks



from functools import partial
import numpy as onp
#from jax import numpy
#from jax.config import config as jax_config
from pyscf import numpy
from pyscf.lib import ops
from .jax_helper import jit
#jax_config.update("jax_enable_x64", True)

einsum = numpy.einsum
dot = numpy.dot

__all__ = ['numpy', 'einsum', 'dot',
           'PLAIN', 'HERMITIAN', 'ANTIHERMI', 'SYMMETRIC',
           'unpack_triu', 'unpack_tril',]

PLAIN = 0
HERMITIAN = 1
ANTIHERMI = 2
SYMMETRIC = 3

@partial(jit, static_argnums=1)
def unpack_triu(triu, filltril=PLAIN):
    '''
    Unpack the upper triangular part of a matrix
    '''
    assert triu.ndim == 1
    nd = int(onp.sqrt(2*triu.size))
    shape = (nd, nd)
    out = numpy.zeros(shape, dtype=triu.dtype)
    idx = onp.triu_indices(nd)
    out = ops.index_update(out, idx, triu)
    if filltril == PLAIN:
        return out
    elif filltril == HERMITIAN:
        out = out + out.conj().T
        out = ops.index_mul(out, onp.diag_indices(nd), .5)
        return out
    elif filltril == ANTIHERMI:
        out = out - out.conj().T
        return out
    elif filltril == SYMMETRIC:
        out = out + out.T
        out = ops.index_mul(out, onp.diag_indices(nd), .5)
        return out
    else:
        raise KeyError

@partial(jit, static_argnums=1)
def unpack_tril(tril, filltriu=PLAIN):
    '''
    Unpack the lower triangular part of a matrix
    '''
    assert tril.ndim == 1
    nd = int(onp.sqrt(2*tril.size))
    shape = (nd, nd)
    out = numpy.zeros(shape, dtype=tril.dtype)
    idx = onp.tril_indices(nd)
    out = ops.index_update(out, idx, tril)
    if filltriu == PLAIN:
        return out
    elif filltriu == HERMITIAN:
        out = out + out.conj().T
        out = ops.index_mul(out, onp.diag_indices(nd), .5)
        return out
    elif filltriu == ANTIHERMI:
        out = out - out.conj().T
        return out
    elif filltriu == SYMMETRIC:
        out = out + out.T
        out = ops.index_mul(out, onp.diag_indices(nd), .5)
        return out
    else:
        raise KeyError



import inspect
import jax
from jaxopt import linear_solve
from jaxopt.tree_util import tree_sub, tree_scalar_mul


def _map_back(diff_items, items, keys):
    new_items = list(items)
    for k, i in enumerate(keys):
        new_items[i] = diff_items[k]
    return tuple(new_items)

def root_vjp(optimality_fun, sol, args, cotangent,
             solve=linear_solve.solve_gmres, nondiff_argnums=()):

    def fun_sol(sol):
        return optimality_fun(sol, *args)

    _, vjp_fun_sol = jax.vjp(fun_sol, sol)

    matvec = lambda u: vjp_fun_sol(u)[0]
    v = tree_scalar_mul(-1, cotangent)
    u = solve(matvec, v)

    diff_args_dict = {i: arg for i, arg in enumerate(args) if i+1 not in nondiff_argnums}
    keys = diff_args_dict.keys()
    diff_args = tuple(diff_args_dict.values())
    def fun_args(*diff_args):
        new_args = _map_back(diff_args, args, keys)
        return optimality_fun(sol, *new_args)

    _, vjp_fun_args = jax.vjp(fun_args, *diff_args)
    diff_vjps = vjp_fun_args(u)

    vjps = [None,] * len(args)
    vjps = _map_back(diff_vjps, vjps, keys)
    return vjps

def _custom_root(solver_fun, optimality_fun, solve,
                 has_aux=False, nondiff_argnums=(), use_converged_args=None):
    solver_fun_sig = inspect.signature(solver_fun)
    optimality_fun_sig = inspect.signature(optimality_fun)

    def make_custom_vjp_solver_fun(solver_fun, kwargs):
        @jax.custom_vjp
        def solver_fun_close_kwargs(*args):
            return solver_fun(*args, **kwargs)

        def solver_fun_fwd(*args):
            res = solver_fun_close_kwargs(*args)
            return res, (res, args)

        def solver_fun_rev(tup, cotangent):
            res, args = tup
            if has_aux:
                cotangent = cotangent[0]
                sol = res[0]
            else:
                sol = res

            args = list(args)
            if use_converged_args:
                for key, val in use_converged_args.items():
                    args[key] = res[val]

            try:
                _ = optimality_fun_sig.bind(*args)
            except TypeError as err:
                msg = ("The optimality function has arguments that "
                       "are not compatible with the solver function.")
                raise TypeError(msg) from err

            vjps = root_vjp(optimality_fun, sol,
                            args[1:], cotangent, solve=solve,
                            nondiff_argnums=nondiff_argnums)
            vjps = (None,) + vjps
            return vjps

        solver_fun_close_kwargs.defvjp(solver_fun_fwd, solver_fun_rev)
        return solver_fun_close_kwargs

    def wrapped_solver_fun(*args, **kwargs):
        ba = solver_fun_sig.bind(*args, **kwargs)
        ba.apply_defaults()
        return make_custom_vjp_solver_fun(solver_fun, ba.kwargs)(*ba.args)

    return wrapped_solver_fun

def custom_root(optimality_fun, solve=None, has_aux=False,
                nondiff_argnums=(), use_converged_args=None):
    if solve is None:
        solve = linear_solve.solve_gmres

    def wrapper(solver_fun):
        return _custom_root(solver_fun, optimality_fun, solve,
                            has_aux=has_aux, nondiff_argnums=nondiff_argnums,
                            use_converged_args=use_converged_args)

    return wrapper

def custom_fixed_point(fixed_point_fun, solve=None, has_aux=False,
                       nondiff_argnums=(), use_converged_args=None):

    def optimality_fun(x0, *args):
        return tree_sub(fixed_point_fun(x0, *args), x0)

    optimality_fun.__wrapped__ = fixed_point_fun

    return custom_root(optimality_fun, solve=solve,
                       has_aux=has_aux, nondiff_argnums=nondiff_argnums,
                       use_converged_args=use_converged_args)
